// $Author: jmartin $
// $Date: 2000-02-05 02:18:35 $
// $Source: /root/cvs/Rigi/rigiedit/UT/src/ut_telos.C,v $
// $Locker:  $
// $State: Exp $
//
// ------------------------------------------------------------------------

#include "ut_telos.h"		// declarations for this file
#include "tml.H"		// for TELOS/mbus stuff
#include "sexps.h"		// read/write s-expressions

#include <unistd.h>		// for usleep()
#include <stdarg.h>		// for varargs stuff
#include <ctype.h>		// for isspace()

/*
	RSF-ish string-based interface, Rigi-UT. This is the UT side;
	it is compiled with the source code provided by UT to make an
	interface library, to be linked with the Rigi editor.

	Info is received/sent to/from rigiedit in RSF, ie. triples
	of the form <relnname> <object> <other_object_or_value>.
	It is the job of this module to perform the conversion to
	TELOS-based NetworkObjects, and to read/write these objects
	to the data server, via mbus.

	[941019] In order to make SEMANTICALLY valid s-expressions, the 
	schema of the DB must be adhered to. The schema is definitely a
	UT issue, so this stuff is taken care of here. For now, we just
	"hardcode" the schema, listing valid arctypes (PROPOSITION). It
	is the responsibility of the Rigi Domain Model to ensure that
	the classes of objects and their attributes are valid.
*/


// The following are schema "arc" relations (between objects), as opposed
// to schema "attribute" relations (between object and value).
// We need to hack it so that we can recognize the PLAS relations
// between PLAS objects. The name space is completely different except
// that the PLAS relations use contains and possibly composite.

// NOTE: if this file knew about the current domain, these hacks
// wouldn't be necessary! What is the best solution??

static const int NUMSCHEMAARCS = 8;
static char* schema_arcs[NUMSCHEMAARCS] =
    { "call", "calls", "composite", "contains", "data", "references", "rigicontains", "struct" };

static void	Error( char*, ... );
static void	PrintErrorMessages( Response* );

static int	InitRigiClient( char*, int );
static int	CloseRigiClient();
static void	SendTellData();

static void	SendRequest( tml_Method*, char* );
static void	FlushRequest();

static int	WaitForResponse(int);
static void	ReceiveResponse();
static int	ProcessResponse();

static const char*	GetNextSexp();
static const char*	GetRequestId( Request* );

static int	SetNextAttribute( char*,char*,char* );	// RSF style
static void	CloseAttributes();
static void	AddSexp( const char* );
static void	AddArgs( Request *r, char *args );

static char arcbuff[256], srcbuff[256];

static const int USLEEPINTERVAL = 100000;	// microseconds
static int WAIT_TIME = 10;			// seconds
static int MAXOBJECTS = 40;

static int server_init = 0;
static int tellinit = 0;

static Request *current_request = NULL;
static Response *current_response = NULL;
static ReturnValueList *rvl = NULL;
static char *current_id = NULL;

// The read_callback function pointer is used by the server to send back
// information in RSF tuple form. It is called from within parse_callback.

static int read_callback_calls;
static void (*read_callback)( char*,char*,char* ) = NULL;
static void parse_callback( char *s, int d, int i, int t );

static int ut_debug = 0;


// These darn UT_interface functions need some updating! Soon.

int UT_interface( int op, char*a, char*b, char*c )
{
    switch (op)
    {
      case UT_ASK:

	Error("error, improper ASK call");
	break;

      case UT_TELL:

	// Add an attribute to current NetworkObject
	if (!server_init)
	{
	    Error("error, TELL attempted before START");
	    break;
	}

	tellinit = 1;

	// write the info to the current sexp
	SetNextAttribute( a,b,c );
	break;

      case UT_START:
      {

	if (server_init) Error("warning, server is being re-initialized");

	char *server = a;
	int port = b ? atoi(b) : -1;
	server_init = InitRigiClient( server, port );
	break;
      }

      case UT_STOP:
      {

	if (!server_init)
	    Error("warning, connection being closed when already so");
	else if (!CloseRigiClient())
	    Error("error, can't close connection");

	server_init = 0;
	break;
      }

      case UT_END:

	if (!server_init) { Error("error, END attempted before START"); break; }

	if (tellinit) SendTellData();

	tellinit = 0;
	break;

      case UT_TELOS:

	// Generic Telos commands
	if (!server_init)
	    { Error("error, COMMAND attempted before START"); break; }

	if (!RESPONSE_QUEUE->empty())
	    Error("warning, response queue not empty!");
	SendRequest( TDS_COMMAND, a );

	if (!WaitForResponse(WAIT_TIME))
	    Error("error, didn't get a response for a COMMAND request");

	else ProcessResponse();
	current_request = NULL;

        break;

      case UT_DEBUG:

	ut_debug = !ut_debug;
	Error("debug flag set %s",ut_debug?"on":"off" );
	break;

      case UT_TIME:
      {
	int i = WAIT_TIME;
	WAIT_TIME = atoi(a);
	Error("wait time set to %d s (formerly %d)",WAIT_TIME,i);
	break;
      }
      case UT_MAXOBJECTS:

	op = atoi(a);
	Error("max objects per message set to %d (formerly %d)",
		op,MAXOBJECTS);
	MAXOBJECTS = op;
	break;

      case UT_FLUSH:
      {
	int i;
	for (i=0; WaitForResponse(0); i++ );
	Error("%d message%s flushed",i,i==1?"":"s");
	break;
      }
      default:

	Error("error, bogus parameter %d in UT_interface()",op);
	break;
    }
    return 1;	// success
}

int UT_interface( int op )
{
    return UT_interface( op, NULL,NULL,NULL );
}

// With this function, the client (Rigi) tells this interface library
// what it wants (s), and what function to use to return the info
// in RSF form (f).

int UT_interface( int op, char *s, void (*f)(char*,char*,char*) )
{
    if ( op!=UT_ASK )
    {
	Error("error, improper ASK use of UT_interface()");
	return 0;
    }
    if (!s)
    {
	Error("error, UT_interface: nothing ASKed for!");
	return 0;
    }

    if (!server_init)
    {
	Error("error, UT_interface not yet STARTed!");
	return 0;
    }

    read_callback_calls = 0;
    read_callback = f;

    // Build and send a request message to the server
    if (!RESPONSE_QUEUE->empty())
	Error("warning, response queue not empty!");

    SendRequest( TDS_ASK, s );

    // Wait for response(s) from server
    for ( int i=0; ;i++ )
    {
	if (!WaitForResponse(WAIT_TIME))
	{
	    Error("error, only %d response%s for \"%s\", no terminator",
		i, i==1?"":"s", s );
	    return 0;
	}

	// Now, get every s-expr out of the RESPONSE object.
	// Break out of the loop if no more Responses are expected
	// (the usual case, but PMDB (for example) may send
	// multiple responses via the "@continue" mechanism).

	if (!ProcessResponse()) break;
    }
    current_request = NULL;
    return read_callback_calls;
}

static int ProcessResponse()
{
    const char *p; int flag = 0;

    // Be getting all the messages out of the message

    while ( p = GetNextSexp() )
    {
	if (flag) Error( "error, too many returnvalues in Response" );

	// PMDB code for indicating that there be more Responses
	flag = !strcmp( p, "@continue" );

	if (!flag)
	{
	    if (ut_debug) Error("Parsing sexp %s",p);
	    parse_sexp( p,parse_callback );
	}
    }
    return flag;
}

static int InitRigiClient( char *server, int port )
{
    Command *C; Boolean r;
    static int tml_init = 0;

    // Have to remake those internal TML things. Mem leak if more than once
    TMLinit();

    CLIENT->setToolClass(RIGI);

    // starting the server with default values
    // i.e., port number 5555 & the same host as client

    if (server && *server)
    {
	String str(server);
	Error( "setting SERVER host id to \"%s\"", server );
	SERVER->setHostId( &str );
    }

    if (port>=0)
    {
	Error( "setting SERVER port number to %d",port );
	SERVER->setPortNumber( port );
    }

    C = new Command(MS_OPEN);
    r = C->sendCommand();
    delete C;
    if (!r)
    {
	String *str = SERVER->getHostId();
        Error("error, can't connect to server on host %s at port %d",
            str->StringToChar(),port);
        return 0;
    }

    // accept message intended for the tool RIGI

    CLIENT->addDomain(RIGI);

    // If this command is not sent, then the client cannot
    // receive any messages.

    C = new Command(MS_ACCEPT_DOMAINS);
    r = C->sendCommand();
    delete C;
    if (!r)
    {
        Error("error, can't command server to accept domains");
        return 0;
    }

    CLIENT->print();
    SERVER->print();
    return 1;
}

static int CloseRigiClient()
{
    Command *C; Boolean r;
    C = new Command(MS_CLOSE);
    r = C->sendCommand();
    delete C;
    if (!r)
    {
        Error("error, can't close connection to server");
        return 0;
    }
    return 1;
}

static void SendTellData()
{
    // flush last pending sexpr to the current request message
    char *p = get_sexp();
    if (p)
    {
	CloseAttributes();
	p = get_sexp();
	AddSexp(p);
	reset_sexp();
    }
    else Error("error, no sexp to flush; no argument added to request");

    if (!RESPONSE_QUEUE->empty())
	Error("warning, response queue not empty!");

    FlushRequest();
    if (!WaitForResponse(WAIT_TIME))
	Error("error, didn't get a response for a TELOS tell");

    current_request = NULL;
}

static void AddSexp( const char *s )
{
    static int i=0;

    if (!current_request)
    {
	current_request = new Request( TELOS_DATA_SERVER );
	current_request->setMethod(TDS_TELL);
	i = 0;
    }
    i++;
    if (ut_debug) Error("adding argument %d:\n%s",i,s); 
    current_request->addArgument((char*)s);

    // I not be thinking that these TML/mbus messages have infinite
    // capacity, so we send info in chunks.

    if (i==MAXOBJECTS)
    {
        if (!RESPONSE_QUEUE->empty())
	    Error("warning, response queue not empty!");

	FlushRequest();
	if (!WaitForResponse(WAIT_TIME))
	    Error("error, didn't get a response for a TELOS tell");
	current_request = NULL;
    }
}

static void FlushRequest()
{
    static int i=0;
    i++;
    Error("telling request #%d",i);

    if (!current_request)
    {
	Error("error, unexpected NULL tell request to flush");
	return;
    }

    const char *id = GetRequestId( current_request );
    Error("TELL Request %s",id);

    if (current_id) free(current_id);
    current_id = strdup(id);

    Boolean r = current_request->sendRequest();
    if (!r) Error("error, couldn't send request!");
    delete current_request;
    current_request = NULL;
}

static void SendRequest( tml_Method *Method, char *s )
{
    // For handling general queries to the server, s could contain
    // multiple substrings, eg. "getInstances RigiObject".
    // Method: TDS_ASK, TDS_TELL, TDS_COMMAND

    char *name;
    if (Method==TDS_ASK) name = "ASK";
    else if (Method==TDS_TELL) name = "TELL";
    else if (Method==TDS_COMMAND) name = "COMMAND";

    Request *request = new Request(TELOS_DATA_SERVER);
    request->setMethod( Method );

    AddArgs( request, s );

    const char *id = GetRequestId( request );
    Error("%s Request %s, string \"%s\"",name,id,s);

    if (current_id) free(current_id);
    current_id = strdup(id);

    Boolean r = request->sendRequest();
    if (!r) Error("error, couldn't send %s request!",name);
    delete request;
}


// In this routine, the Rigi client waits (awhile) for the next RESPONSE
// message. Rigi SHOULD be expecting this response,
// with matching id, in response to some previous request.
// Matching Request should still be in existence.
// Wait for only j seconds.

static int WaitForResponse( int j )
{
    int i=0;
    NewMessageStatus status;

    j *= 1000000;	// convert to microseconds

    for ( i=0; i<=j; i+=USLEEPINTERVAL )
    {
        usleep( USLEEPINTERVAL );

        status = TMBgetNewMessage();

        switch (status)
	{
            case NO_NEW_MESSAGE:
                break;
            case NEW_RESPONSE:
		ReceiveResponse();
                break;
            case NEW_REQUEST:
            case UNRECOGNIZABLE_MESSAGE:
            case NEW_MONOLOGUE:
            default:
                Error("Hey! Bogus message! (status==%d)",status);
                break;
        }
        if (status!=NO_NEW_MESSAGE) break;
    }
    if ( status!=NEW_RESPONSE ) return 0;

    return 1;
}

static void ReceiveResponse()
{
    if (current_response) delete current_response;
    Response *r = current_response = RESPONSE_QUEUE->front();

    rvl = r->getReturnValues();

    String *str = r->getResponseTo();
    const char *s = str ? str->StringToChar() : "<null>";

    Error( "got a Response for %s", s );

    if (strcmp(current_id,s))
	Error( "error, mismatched id, expected %s", current_id );

    if (r->getCompletionStatus()!=SUCCESS)
    {
	int n = r->getErrorMessages()->getSize();
	Error( "error, bad Response received (code %d), %d message%s%s",
	    r->getCompletionStatus(), n, n==1?"":"s", n==0?"":":" );
	PrintErrorMessages( r );
    }

    if (!rvl)
	Error("error, no returnvaluelist!");
    else if (ut_debug)
	Error("number of returnvalues is %d", rvl->getSize() );

    RESPONSE_QUEUE->leave();
}


    // Here be an s-expr parse callback function. The syntax of
    // NetworkObject s-exprs is implied by the code herein.
    // (If the syntax for s-exprs that we expect to receive is
    // any more complex, we couldn't use this approach!)

    static void parse_callback( char *s, int d, int i, int top )
    {
	if (!read_callback) Error("error, problem in parse_callback");
	//if (ut_debug) printf("parse_callback: %d %d %d %s\n",d,i,t,s );

	if (d==0) if (i==0) strcpy( srcbuff,s );
	if (d==1) if (strcmp(s,"Token"))		// HACK!
	{
	    read_callback_calls++;
	    if (top==2)
		(*read_callback)( "type",srcbuff,s );	// HACK!
	    else if (top==3)
		(*read_callback)( "isa",srcbuff,s );	// HACK!
	    else
		Error("error, bad syntax in s-exp on token %s",s);
	}
	if (d==3) if (i==0) strcpy( arcbuff,s );
	if (d==4) if (i==0)
	{
	    read_callback_calls++;
	    (*read_callback)( arcbuff, srcbuff, s );
	}
    }


// This routine returns the next s-expr contained in the current RESPONSE
// object. It returns NULL if there are no more.

static const char *GetNextSexp()
{
    static int i=0;
    ReturnValueString *rvs;

    if (!rvl)
    {
	Error("error, no returnvaluelist available from response!");
	return NULL;
    }
    rvs = (*rvl)[i++];
    if (!rvs) { i=0; return NULL; }

    const char *p = rvs->StringToChar();
    return p;
}

static const char *GetRequestId( Request *r )
{
    if (!r)
    {
	Error("error, attempt to get ID of null Request!");
	return NULL;
    }

    String *str = r->getId();
    return str ? str->StringToChar() : "<null>";
}

static int IsSchemaArc( char *arc )
{
    for ( int i=0; i<NUMSCHEMAARCS; i++ )
	if (!strcmp(arc,schema_arcs[i])) return 1;
    return 0;
}

static char *objname = NULL;
static char *arcname = NULL;

static int SetNextAttribute( char *arc, char *src, char *dst )
{
    // Here, we add info from a new tuple to the current s-expr,
    // or to a new s-expr. We must insist that the first RSF
    // triple of info must be a "type" relation.
    // This unfortunate hack is convenient because the "type"
    // of an object ("class" in Telos terms) occurs at the
    // beginning of an s-expression. HACK - should be fixed.

    static int isproposition;

    //printf("next attribute:     %s %s %s\n",arc,src,dst );

    if ( !objname || strcmp(src,objname))
    {
	// This is a new object name, so flush last sexp to request
	char *p = get_sexp();
	if (p)
	{
	    CloseAttributes();
	    p = get_sexp();
	    AddSexp(p);
	    reset_sexp();
	}

	// Oops, I've strayed from %*!^%$ C++

	if (arcname) free(arcname);
	arcname = NULL;

	if (objname) free(objname);
	objname = strdup(src);

	char *type = "";
	if (strcmp(arc,"type"))
	    Error("error, first RSF triple for %s must be \"type\"",src);
	else
	    type = dst;
        reset_sexp();

        write_sexp("( ");
        write_sexp( src );
	write_sexp(" Token (");
	write_sexp( type );
	write_sexp(") () (\n");

	if (type==dst) return 1;
    }

    // Do we have a new attribute category name? 
    if ( !arcname || strcmp(arcname,arc) )
    {
	if (arcname)
	{
	    // close off the attributevalue list for the last arcname
	    write_sexp(" ) )\n");
	    free( arcname );
	}
	arcname = strdup(arc);
	isproposition = IsSchemaArc( arcname );

	write_sexp("\t( ( ");
	write_sexp( arc );
	write_sexp(" ) (");
    }

    // Finally, write the dst object's name
    write_sexp(" ( ");
    if (!isproposition) write_sexp("\"");
    write_sexp( dst );
    if (!isproposition) write_sexp("\"");
    //write_sexp( isproposition?" PROPOSITION":" STRING");
    write_sexp(" )");

    return 1;
}

static void CloseAttributes()
{ 
    // Close off an attr category list, if it exists
    if (arcname) { write_sexp(" ) )"); free( arcname ); }

    // Finish off the whole s-epr for the current object
    write_sexp(" ) )");
    arcname = NULL;

    free( objname ); objname = NULL;
}

static void AddArgs( Request *r, char *args )
{
    // Here, args might be a list of names, separated by white space

    for (;;)
    {
	while (*args && isspace(*args)) args++;
	if (!*args) break;
	char *p = index( args,' ' );
	if (p) *p = '\0';
	r->addArgument(args);
	if (p) *p = ' '; else break;
	args = p+1;
    }
}

static void PrintErrorMessages( Response *r )
{
    String *str;

    int i=0;
    while ( str = r->getErrorMessages()->removeErrorMessage(0) )
    {
	const char *s = str->StringToChar();
	fprintf( stderr,"%2d: %s\n",i,s );
	i++;
    }
    fflush(stderr);
}

static void Error( char *format ... )
{
    va_list ap;
    va_start( ap, format );
    fprintf( stderr,"Rigi/mbus: " );
    if (format) vfprintf( stderr, format, ap );
    fprintf( stderr,"\n" );
    va_end(ap);
}

// ------------------------------------------------------------------------
//
// $Log: not supported by cvs2svn $
// Revision 1.1  1997/12/05 01:57:18  jmartin
// Initial revision
//
//   Revision 1.6  1995/10/30  22:12:37  mwhitney
//   Changes supporting COMMAND requests, incl. removal
//   of old "HACK" commands
//
//   Revision 1.5  1995/10/27  00:52:06  mwhitney
//   many changes
//
//   Revision 1.4  1995/02/15  19:31:38  mwhitney
//   Many changes!
//
//   Revision 1.3  1994/10/28  18:07:01  mwhitney
//   many many changes! Panic for CASCON94.
//
//   Revision 1.2  1994/09/12  22:00:17  mwhitney
//   Many changes
//
//   Revision 1.1  1994/08/17  20:48:10  mwhitney
//   Initial revision
//
//   Revision 1.2  1994/08/02  17:55:41  mwhitney
//   introduced use of ut.h
//
//   Revision 1.1  1994/07/12  16:53:06  mwhitney
//   Initial revision
//
