#include "String.H"
extern "C" {
#include "api.h"
}
#include "mbus-interface.H"
#include "os.H"

#define MBUS_PORT_NUMBER             5555
#define MBUS_HOSTNAME              "luna.ai.toronto.edu"
    //  default port number
#define PROCESS_ID_LENGTH          5    
    //  process id should be 5 digit number
#define MESSAGE_ID_LENGTH        256    
    //  the maximum number of characters allowed for a message id
#define MAX_DOMAIN_LENGTH      10000    
    //  the maximum number of characters allowed for a MBUS domain string

#define TRANS_TAG_POSITION          0
#define TRANS_DOMAINS_POSITION      1
#define TRANS_DATA_POSITION         2

#define MESSAGE_ID_POSITION         0
#define MESSAGE_SENDER_POSITION     1
#define MESSAGE_RECEIVER_POSITION   2
#define MESSAGE_TYPE_POSITION       3
#define MESSAGE_METHOD_POSITION     4
#define MESSAGE_ARGUMENTS_POSITION  5

#define MESSAGE_RESPONSETO_POSITION 4
#define MESSAGE_STATUS_POSITION     5
#define MESSAGE_RETVALUES_POSITION  6
#define MESSAGE_ERRMSGS_POSITION    7

#define MESSAGE_MONOLOGUE_POSITION  4

#define DOMAIN_CONTEXT_POSITION     0
#define DOMAIN_TOOL_POSITION        1
#define DOMAIN_USERID_POSITION      2


enum MessageType { MESSAGE, MONOLOGUE, REQUEST, RESPONSE, COMMAND };
    // MONOLOGUE similar to a request except that there is no 
    //           response expected by the sender of the message
    // REQUEST   a message that is intended to be received, and 
    //           responded to, by one or more clients 
    // RESPONSE  a message that is intended to be received as a 
    //           response to a request by a client or a command 
    //           by the message server
    // COMMAND   a message intended to be received, and responded 
    //           to, by the message server itself


enum NewMessageStatus { NO_NEW_MESSAGE, UNRECOGNIZABLE_MESSAGE,
                        NEW_REQUEST, NEW_RESPONSE, NEW_MONOLOGUE };


// commands accepted by message server (prefix 2)
// ----------------------------------------------
#define MS_OPEN               201
    //    Connects to the message server on the port and host specified in 
    //    SERVER. The userId specified in the CLIENT (based on the login name, 
    //    ToolClass, clientHostId and clientProcessId) is registered with the
    //    message server. 
#define MS_CLOSE              202
    //    Closes the connection to the message server. When the command
    //    is executed, it first ensures that all previous messages associated 
    //    with the client have been received and processed by the message 
    //    server.
#define MS_CLEAR              203
    //    Clears all expressed interest in classes of messages defined by
    //    ToolClass and Context. After a MS_CLEAR, a client will receive
    //    only messages that specifically specify the client userId.
#define MS_ACCEPT_DOMAINS     204
    //    Accepts the userIds, ToolClasses, or Contexts (listed in the
    //    DomainList of the client) as the client's domain of interest.
#define MS_GET_CLIENTS        205 
    //    Returns a list of the userIds of all the current clients.


#define SUCCESS_CHAR    '0'
#define UNKNOWN_CHAR    '1'
#define ERROR_CHAR      '2'
#define NULLSTATUS_CHAR '3'

enum CompletionStatus { SUCCESS, UNKNOWN, ERROR, NULLSTATUS };
    // SUCCESS    indicating the successful completion of the request
    // UNKNOWN    indicating the inability of the responding client to 
    //            handle the request
    // ERROR      indicating the failure to handle the request
    // NULLSTATUS indicating default value during initialization

CompletionStatus
t_sexprToCompletionStatus(t_sexp sexp);


// GenericInteger allows us to enter integers into generic lists.

class GenericInteger : public Generic 
{  public:
    int value;
    GenericInteger(int i) { value = i; };
};

class GenericString: public Generic, public String {
  public:
    GenericString(String s) : String(s) {};
    GenericString(char* s)  : String(s) {};
    GenericString()         : String()  {};
};

// MessageIdString type is used to uniquely identifies each message. 
// The id of a message is generated by the library. The id is primarily
// used for bookkeeping purposes (eg. to identify corresponding requests
// and responses. It becomes even more important for asynchronous message
// passsing.

class MessageIdString : public Generic, public String 
{
  public:
    MessageIdString(String s)     : String(s) {};
    MessageIdString(char* s)      : String(s) {};
    MessageIdString()             : String()  {};
};


// MonologueString type is used to store the actual contents of the 
// monologue message. This kind of message could be useful to support
// a mailing facility, as well as, a broadcasting facility to communicate
// new releases, updates, etc.,.

class MonologueString : public Generic, public  String 
{
  public:
    MonologueString(String s) : String(s) {};
    MonologueString(char* s)  : String(s) {};
    MonologueString()         : String()  {};
};


// ArgumentString type is used to store an argument to either a method in a 
// request or a command in a command message.

class ArgumentString : public Generic, public  String 
{
  public:
    ArgumentString(String s) : String(s) {};
    ArgumentString(char* s)  : String(s) {};
    ArgumentString()         : String()  {};
};


// ArgumentList is used to store the arguments of a method in a request or
// a command in a command message.

class ArgumentList : public GenericList
{
public:
  ArgumentList(): GenericList() {};
  ArgumentList(t_sexp sexp);

    // read only routines
    ArgumentString* getArgument(int i) 
                        { return (ArgumentString*)getItem(i); };
    int findArgument(ArgumentString* s) { return findItem(s); };

    ArgumentString* operator[] (int index) const
        { if ((index > -1) && (index < this->getSize()))
             return (ArgumentString*)this->getItem(index);
          else
             return (ArgumentString*)NULL;
        }

    void print (ostream& outStream = cout);

    // write access routines
    void addArgument(ArgumentString* s)    
      { 
        ArgumentString* arg = new ArgumentString(*s);
        addItem(arg);      
      };

    void addArgument(char* s)    
      { 
        ArgumentString* arg = new ArgumentString(s);
        addItem(arg);      
      };

    void removeArgument(ArgumentString* s) { removeItem(s);   };
    void clearArgumentList()               { clear();         };

    ArgumentString* removeArgument(int i)  
                   { return (ArgumentString*)removeItem(i);   };
};



// ContextString type is used to describe domains of interests for the sender
// or the receiver of a message. The concept of context is underdeveloped as
// of yet, but will evolve throughout the length of the project.

class ContextString : public Generic, public  String 
{
  public:
    ContextString(String s) : String(s) {};
    ContextString(char* s)  : String(s) {};
    ContextString()         : String()  {};
};



// ToolClass type is used to define the different kinds of tools that the
// environment supports. Currently it supports a message server, a repository,
// and two different reverse engineering tools, REFINE and RIGI.

class ToolClass : public Generic, public String
{
  public:
    ToolClass(String s) : String(s) {};
    ToolClass(char* s)  : String(s) {};
    ToolClass()         : String()  {};
};



extern ToolClass* TELOS_DATA_SERVER;
    // indicating the current Telos data server
extern ToolClass* REFINE;
    // indicating a Refine tool  client
extern ToolClass* RIGI;
    // indicating a Rigi tool client
extern ToolClass* MESSAGE_SERVER;
    // indicating the current message server



// tml_Method is a method exported from (or supported by) a tool.
class tml_Method : public Generic, public String
{
  public:
    tml_Method(String s) : String(s) {};
    tml_Method(char* s)  : String(s) {};
    tml_Method()         : String()  {};
};


// extern tml_Method *TDS_TELL;
// extern tml_Method *TDS_ASK;
// extern tml_Method *TDS_COMMAND;

// MessageType is used to distinguish the four different kinds of messages
// supported by the current environment: a monologue, a request, a response 
// and a command.


    // enum MessageType { MESSAGE, MONOLOGUE, REQUEST, RESPONSE, COMMAND };
    // MONOLOGUE similar to a request except that there is no 
    //           response expected by the sender of the message
    // REQUEST   a message that is intended to be received, and 
    //           responded to, by one or more clients 
    // RESPONSE  a message that is intended to be received as a 
    //           response to a request by a client or a command 
    //           by the message server
    // COMMAND   a message intended to be received, and responded 
    //           to, by the message server itself


// UserIdString is used to uniquely identify the specific user/tool within 
// the environment. There is a user id issued for each running client and
// is a combination of the login name, tool class, host id and process number.

class UserIdString : public Generic, public String 
{
  public:
    UserIdString(String s)       : String(s) {};
    UserIdString(const char* s)  : String(s) {};
    UserIdString()               : String()  {};
};



// ReturnValueString is used to return a value in a response to a 
// request/command.  
// 
class ReturnValueString : public Generic, public  String 
{
  public:
    ReturnValueString(String s) : String(s) {};
    ReturnValueString(const char* s)  : String(s) {};
    ReturnValueString()         : String()  {};
};



// The Domain type is used by a user/client to specify the kinds of messages 
// that it is interested in. It is also used by the sender to specify the 
// receiver(s) of a message. Domain is a tuple consisting of
// (a context, a tool class, a user id).

class Domain : public Generic
{   ContextString* context;
    ToolClass*     tool;
    UserIdString*  userId;
  public:
    Domain() 
      { context = new ContextString();
        tool = new ToolClass();
        userId = new UserIdString();
      };

    Domain(ContextString* c) 
      { context = new ContextString(*c);
        tool = new ToolClass();
        userId = new UserIdString();
      };

    Domain(ToolClass* t) 
      { context = new ContextString();
        tool = new ToolClass(*t);
        userId = new UserIdString();
      };

    Domain(UserIdString* u) 
      { context = new ContextString();
        tool = new ToolClass();
        userId = new UserIdString(*u);
      };

    Domain(ContextString* c, ToolClass* t) 
      { context = new ContextString (*c);
        tool = new ToolClass(*t);
        userId = new UserIdString();
      };

    Domain(ContextString* c, UserIdString* u) 
      { context = new ContextString(*c);
        tool = new ToolClass();
        userId = new UserIdString(*u);
      };

    Domain(ToolClass* t, UserIdString* u) 
      { context = new ContextString();
        tool = new ToolClass(*t);
        userId = new UserIdString(*u);
      };

    Domain(ContextString* c, ToolClass* t, UserIdString* u) 
      { context = new ContextString(*c);
        tool = new ToolClass(*t);
        userId = new UserIdString(*u);
      };

    Domain(t_sexp sexp)
      { t_sexp c = MBcar(MBnth(sexp, DOMAIN_CONTEXT_POSITION));
        t_sexp t = MBcar(MBnth(sexp, DOMAIN_TOOL_POSITION));
        t_sexp u = MBcar(MBnth(sexp, DOMAIN_USERID_POSITION));
        if (c == (t_sexp)NULL)
           context = new ContextString();
        else
	  { char* contextStr = MBprint_Cstring(c);
            context = new ContextString(contextStr);
            (*MBfree_func)(contextStr);
          }

        if (t == (t_sexp)NULL)
           tool = new ToolClass();
        else
	  { char* toolStr =    MBprint_Cstring(t);
            tool = new ToolClass(toolStr);
            (*MBfree_func)(toolStr);
          }

        if (u == (t_sexp)NULL)
           userId = new UserIdString();
        else
	  { char* userIdStr =  MBprint_Cstring(u);
            userId = new UserIdString(userIdStr);
            (*MBfree_func)(userIdStr);
          }
      };

    ~Domain()
       { delete(context);
         delete(tool);
         delete(userId);
       };

    UserIdString*  getUserId()    { return userId;  };
    ToolClass*     getToolClass() { return tool;    };
    ContextString* getContext()   { return context; };

    void print (ostream& outStream = cout)
      { if (context->StringLength() == 0)
	  { if (tool->StringLength() == 0)
               outStream << *userId;
            else
              { if (userId->StringLength() == 0)
                   outStream << *tool;
                else
                   outStream << *tool << "/" << *userId;
     	      };
          }
        else
          { if (tool->StringLength() == 0)
              { if (userId->StringLength() == 0)
                   outStream << *context;
                else
                   outStream << *context << "/" << *userId;
     	      }
            else
              { if (userId->StringLength() == 0)
		  { outStream << *context << "/" << *tool; }
                else
                  { outStream << *context << "/" << *tool << "/" << *userId; }
              }
          }
      };


    void setUserId(UserIdString* i)   { userId  = new UserIdString(*i);  };
    void setToolClass(ToolClass* t)   { tool    = new ToolClass(*t);     };
    void setContext(ContextString* c) { context = new ContextString(*c); };
};



// DomainList is used by a user/client to specify the kinds of messages 
// that it is interested in. It is also used by the sender to specify the 
// receiver(s) of a message. An element of the list is a tuple consisting of
// (a context, a tool class, a user id).

class DomainList : public GenericList
{
public:
  DomainList(): GenericList() {};
  DomainList(t_sexp sexp): GenericList()
      { int n = MBlength(sexp);
        Domain* d;
	for (int i = 0; i < n; i++)
            { d = new Domain(MBnth(sexp, i));
              addItem(d);
            }
      };

    // read only routines
    Domain* getDomain(int i) { return (Domain*)getItem(i); };
    int findDomain(Domain* d){ return findItem(d);         };

    Domain* operator[] (int index) const
        { if ((index > -1) && (index < this->getSize()))
             return (Domain*)this->getItem(index);
          else
             return (Domain*)NULL;
        }

    void print (ostream& outStream = cout)
      { outStream << "(";
        int n = getSize();
        if (n > 0)
	  { for (int i = 0; i < n - 1; i++)
                { getDomain(i)->print(); outStream << ", "; }
            getDomain(n-1)->print();
	   };
        outStream << ")";
      };

    // write access routines
    void addDomain(Domain* d1)    
      { Domain* d2 = new Domain(*d1);                    
        addItem(d2);                    
      };
    void addDomain(UserIdString* u)    
      { Domain* d = new Domain(u);                    
        addItem(d);                    
      };
    void addDomain(ContextString* c)    
      { Domain* d = new Domain(c);                    
        addItem(d);                    
      };
    void addDomain(ToolClass* t)    
      { Domain* d = new Domain(t);                    
        addItem(d);                    
      };

    void removeDomain(Domain* d) { removeItem(d);                 };
    Domain* removeDomain(int i)  { return (Domain*)removeItem(i); };
    void clearDomainList()       { clear();                       };
};

class ReturnValueList : public GenericList
{
public:
  ReturnValueList(): GenericList() {};
  ReturnValueList(t_sexp sexp): GenericList()
      { int n = MBlength(sexp);
        ReturnValueString* arg;
        char* tmp;
	for (int i = 0; i < n; i++)
            { tmp = MBprint_Cstring(MBnth(sexp, i));
              arg = new ReturnValueString(tmp);
              (*MBfree_func)(tmp);
              addItem(arg);
            }
      };

    // read only routines
    ReturnValueString* getReturnValue(int i) 
                           { return (ReturnValueString*)getItem(i); };
    int findReturnValue(ReturnValueString* s) { return findItem(s); };

    ReturnValueString* operator[] (int index) const
        { if ((index > -1) && (index < this->getSize()))
             return (ReturnValueString*)this->getItem(index);
          else
             return (ReturnValueString*)NULL;
        }

    void print (ostream& outStream = cout)
      { outStream << "(";
        int n = getSize();
        if (n > 0)
	  { for (int i = 0; i < n - 1; i++)
                outStream << getReturnValue(i) << ", ";
            outStream << getReturnValue(n-1);
	   };
        outStream << ")";
          
      };

    // write access routines
    void addReturnValue(ReturnValueString* s1)    
      { 
        ReturnValueString* s2 = new ReturnValueString(*s1);
        addItem(s2);      
      };
    void removeReturnValue(ReturnValueString* s) { removeItem(s);   };
    void clearReturnValueList()                  { clear();         };
    ReturnValueString* removeReturnValue(int i)                
                     { return (ReturnValueString*) removeItem(i);   };
};



//////////////////////////////////////////////////////////////////////////////
//                                                                          //
//                                Client                                    //
//                                                                          //
//////////////////////////////////////////////////////////////////////////////


// A Client stores the login name, the host id, the process id, the tool
// class and the domains of interest of the current client.

class Client
{
    String*       loginName;
    ToolClass*    toolClass;
    String*       hostId;
    String*       processId;
    int           processNumber;
    UserIdString* userId;
    String*       messageIdPrefix;
    DomainList*   domains;

    void setUserId() 
      { userId = new UserIdString(*loginName); 
        *userId = *userId + *hostId;
        *userId = *userId + *processId;
      };

    void setMessagePrefix ()
      { messageIdPrefix = new String(*hostId);    
        *messageIdPrefix = *messageIdPrefix + *processId;
      };

  public:
    // constructor
    // -----------
    Client ()
      {  
	domains =   new DomainList();
        loginName = getClientLogin();
        hostId =    getClientHostId();
        toolClass = new ToolClass();
        processId = getClientPId();
        processNumber = getClientPNumber();
        setUserId();
        domains->addDomain(getUserId());
        setMessagePrefix ();
      };

    
    ~Client ()
      {
        delete(loginName);
        delete(toolClass);
        delete(hostId);
        delete(processId);
        delete(userId);
        delete(messageIdPrefix);
        delete(domains);
      };

    // read access to Client data structure
    String*       getLoginName            () { return loginName;              };
    ToolClass*    getToolClass            () { return toolClass;              };
    String*       getHostId               () { return hostId;                 };
    String*       getProcessId            () { return processId;              };
    int           getProcessNumber        () { return processNumber;          };
    UserIdString* getUserId               () { return userId;                 };
    String*       getMessageIdPrefix      () { return messageIdPrefix;        };
    String*       generateMessageIdSuffix () { return getUniqueTimeString();  };
    DomainList*   getDomainList           () { return domains;                };

    void print (ostream& outStream = cout)
      { outStream << endl << "  Client" << endl << "  ------" << endl;
        outStream << "    loginName:        " << *getLoginName() << endl;
        outStream << "    toolClass:        " << *getToolClass() << endl;
        outStream << "    hostId:           " << *getHostId() << endl;
        outStream << "    processId:        " << *getProcessId() << endl;
        outStream << "    userId:           " << *getUserId() << endl;
        outStream << "    messageIdPrefix:  " << *getMessageIdPrefix() << endl;
        outStream << "    domains:          ";
	domains->print();
        outStream << endl;
      };


    void setToolClass (ToolClass* t) { toolClass = new ToolClass(*t);     };
    void addDomain(Domain* d) { domains->addDomain(d);                    };
    void addDomain(UserIdString* u) { domains->addDomain(u);              };
    void addDomain(ContextString* c) { domains->addDomain(c);             };
    void addDomain(ToolClass* t) { domains->addDomain(t);                 };

    void setDomainList(DomainList* d)
      { domains->clear();
        int n = d->getSize();
        for (int i = 0; i < n; i++)
            domains->addDomain(d->getDomain(i));
      };

    void clearDomainList()       { domains->clear();                      };
    void removeDomain(Domain* d) { domains->removeDomain(d);              };
    Domain* removeDomain(int i)  { return domains->removeDomain(i);       };
};



//////////////////////////////////////////////////////////////////////////////
//                                                                          //
//                                Server                                    //
//                                                                          //
//////////////////////////////////////////////////////////////////////////////

class Server
{
    String*         hostId;
    GenericInteger* portNumber;
    GenericInteger* fileDescriptor;
  public:
    // constructor
    // -----------
    Server ()
      {  hostId     = getClientHostId();
         portNumber = new GenericInteger(MBUS_PORT_NUMBER);
         fileDescriptor = new GenericInteger(0);
      };

    ~Server ()
      {
        delete(hostId);
        delete(portNumber);
        delete(fileDescriptor);
      };

    // read access to Server data structure
    String*  getHostId    () { return hostId;                     };
    int getPortNumber     () { return portNumber->value;          };
    int getFileDescriptor () { return fileDescriptor->value;      };

    void print (ostream& outStream = cout)
      { outStream << endl << "  Server" << endl << "  ------" << endl;
        outStream << "    hostId:            " << *getHostId() << endl;
        outStream << "    portNumber:        " << getPortNumber() << endl;
        outStream << "    fileDescriptor:    " << getFileDescriptor() << endl;
      };

    // write accesss to Server data structure
    void setHostId (String* h)       { hostId = new String(*h);   };
    void setPortNumber (int port )   { portNumber->value = port;  };
    void setFileDescriptor (int f)   { fileDescriptor->value = f; };
};  

extern Client *CLIENT;
extern Server *SERVER;


class Message {
    MessageIdString* generateNextMessageId()
      { String* tmp;
        tmp = CLIENT->generateMessageIdSuffix();
        MessageIdString* ret = new MessageIdString(*CLIENT->getMessageIdPrefix() + *tmp);
        delete(tmp);
        return ret;
      };
  public:
    MessageIdString* id;
    Domain* sender;
    DomainList* receiver;

    Message(DomainList* d) 
      { id = generateNextMessageId();
        sender = new Domain;
        sender->setUserId(CLIENT->getUserId());
        sender->setToolClass(CLIENT->getToolClass());
        receiver = new DomainList;

        int n = d->getSize();
        for (int i = 0; i < n; i++)
            receiver->addDomain(d->getDomain(i));
      };

    Message(ContextString* c)
      { id = generateNextMessageId();
        sender = new Domain;
        sender->setUserId(CLIENT->getUserId());
        sender->setToolClass(CLIENT->getToolClass());
        receiver = new DomainList;
        receiver->addDomain(c);
      };

    Message(ToolClass* t)
      { id = generateNextMessageId();
        sender = new Domain;
        sender->setUserId(CLIENT->getUserId());
        sender->setToolClass(CLIENT->getToolClass());
        receiver = new DomainList;
        receiver->addDomain(t);
      };

    Message(UserIdString* u)
      { id = generateNextMessageId();
        sender = new Domain;
        sender->setUserId(CLIENT->getUserId());
        sender->setToolClass(CLIENT->getToolClass());
        receiver = new DomainList;
        receiver->addDomain(u);
      };
    Message()   
      { id = generateNextMessageId();
        sender = new Domain;
        sender->setUserId(CLIENT->getUserId());
        sender->setToolClass(CLIENT->getToolClass());
        receiver = new DomainList;
      };

    ~Message()   
      {
        delete(id);
        delete(sender);
        delete(receiver);
      };

    // read access to Message data structure
    MessageIdString* getId       ()  { return id;            };
    Domain*  getSender           ()  { return sender;        };
    DomainList*   getReceiver    () const { return receiver; };

    void print (ostream& outStream = cout)
      { outStream << "    id:                " << *id     << endl;
        outStream << "    sender:            ";
        sender->print();
        outStream << endl;
        outStream << "    receiver:          ";
        getReceiver()->print();
        outStream << endl;
      };

    // write accesss to Message data structure
    void addReceiver(Domain* d)        { receiver->addDomain(d); };
    void addReceiver(ToolClass* t)     { receiver->addDomain(t); };
    void addReceiver(UserIdString* u)  { receiver->addDomain(u); };
    void addReceiver(ContextString* c) { receiver->addDomain(c); };

    void setReceiver(DomainList* d)
      { receiver->clear();
        int n = d->getSize();
        for (int i = 0; i < n; i++)
            receiver->addDomain(d->getDomain(i));
      };

    void clearReceiverList()      { receiver->clear();                 };
    void removeReceiver(Domain* d){ receiver->removeDomain(d);         };
    Domain* removeReceiver(int i) { return receiver->removeDomain(i);  };

    String* makeMBDomainString(DomainList* domains);
    String* makeMBDomainSpecString(DomainList* domains);
    String* makeDomainString(Domain* d);

     String* makeSenderString()
      { return makeDomainString(getSender()); };

    String* makeDomainListString(DomainList* d);
};

class Monologue;
class Request;
class Response;
class Command;


class Monologue : public Message, public Generic {
    MessageType type;
    MonologueString* monologue;

    const char* getTypeString();

  public:
    Monologue(DomainList* d) : Message(d)  
       { type = MONOLOGUE; 
         monologue = new MonologueString(); };

    Monologue()              : Message()
       { type = MONOLOGUE; 
         monologue = new MonologueString(); };

    Monologue(MonologueString* m)     : Message()   
       { type = MONOLOGUE; 
         monologue = new MonologueString(*m); };

    Monologue(t_sexp sexp);
    ~Monologue()        { delete(monologue); };

    // read access to Message data structure
    MessageType      getType ()         { return type;      };
    MonologueString* getMonologue ()    { return monologue; };

    void print (ostream& outStream = cout);
    String* makeMonologueString();
    Boolean sendMonologue();
};

class Request : public Message, public Generic {

    tml_Method* method;
    ArgumentList* arguments;
    MessageType type;

    const char* getTypeString();

    String* makeArgumentListString();
    String* makeDomainListString(DomainList* d);

     String* makeSenderString()
      { return makeDomainString(getSender()); };

  public:
    // constructor
    // -----------
    Request(DomainList* d) : Message(d)  
      { type = REQUEST;   
        method = new tml_Method();
        arguments = new ArgumentList;
      };

    Request()             : Message()   
      { type = REQUEST;   
        method = new tml_Method();
        arguments = new ArgumentList();
      };

    Request(ContextString* c) : Message(c)
      { type = REQUEST;   
        method = new tml_Method();
        arguments = new ArgumentList;
      };

    Request(ToolClass* t) : Message(t)
      { type = REQUEST;   
        method = new tml_Method();
        arguments = new ArgumentList;
      };

    Request(UserIdString* u) : Message(u)
      { type = REQUEST;   
        method = new tml_Method();
        arguments = new ArgumentList;
      };

    Request(t_sexp sexp);
    ~Request()
      {
        delete(method);
        delete(arguments);
      };

    // read access to Message data structure
    MessageType    getType()                { return type;      };
    tml_Method*   getMethod ()              { return method;    };
    ArgumentList* getArgumentList() const   { return arguments; };

    void
    print(ostream& outStream = cout);
        
    // write access to request data structure
    void setMethod(tml_Method* m)        { method = new tml_Method(*m); };
    void addArgument(ArgumentString* a)  { arguments->addArgument(a);   };
    void addArgument(char* a)            { arguments->addArgument(a);   };

    void setArguments(ArgumentList* a)
      { arguments->clear();
        int n = a->getSize();
        for (int i = 0; i < n; i++)
            arguments->addArgument(a->getArgument(i));
      };

    void clearArgumentList()              { arguments->clear();           };
    ArgumentString* removeArgument(int i) 
                                   { return arguments->removeArgument(i); };
    void removeArgument(ArgumentString* s){ arguments->removeArgument(s); };

    String* makeRequestString();
    Boolean sendRequest() 
      { Boolean ret;
        String* dataStr = makeRequestString();
        String* domainStr  = makeMBDomainSpecString(getReceiver());
        MB_sendMessage(dataStr, domainStr);
        if (TMB_ERROR->found())
           ret = FALSE;
        else
           ret = TRUE;
        delete(dataStr);
        delete(domainStr);
        return ret;
      };
};

class Response : public Message, public Generic {
    MessageType       type;
    MessageIdString*  responseTo;
    CompletionStatus  completion;
    ReturnValueList*  returnValues;
    ErrorMessageList* errorMessages;

    const char* getTypeString();
    const char* getCompletionStatusString();

    String* makeReturnValueListString();
    String* makeErrorMessageListString();

  public:
    // constructor
    // -----------
    Response(DomainList* d) : Message(d)  
      { type = RESPONSE; 
        responseTo = new MessageIdString();
        returnValues  = new ReturnValueList;
        errorMessages = new ErrorMessageList;
      };
    Response()              : Message()
      { type = RESPONSE; 
        responseTo = new MessageIdString();
        returnValues  = new ReturnValueList;
        errorMessages = new ErrorMessageList;
      };

   Response(Request* r) : Message()
  { type = RESPONSE;
    returnValues  = new ReturnValueList;
    errorMessages = new ErrorMessageList;
    responseTo = new MessageIdString(*(r->getId()));
    getReceiver()->addDomain(r->getSender());
  };

    Response(t_sexp sexp);
    ~Response()
      {
        delete(responseTo);
        delete(returnValues);
        delete(errorMessages);
      };

    // read access to Response data structure
    MessageType       getType ()            { return type;          };
    MessageIdString*  getResponseTo()       { return responseTo;    };
    CompletionStatus  getCompletionStatus() { return completion;    };
    ReturnValueList*  getReturnValues()     { return returnValues;  };
    ErrorMessageList* getErrorMessages()    { return errorMessages; };

    void
    print(ostream& outStream = cout);

    // write access to Response data structure
    void setResponseTo(Request* r)
     { responseTo = new MessageIdString(*(r->getId())); };

    void setResponseTo(MessageIdString* s)
     { responseTo = new MessageIdString(*s);                               };
    void setCompletionStatus(CompletionStatus s) { completion = s;         };
    void addReturnValue(ReturnValueString* r)
     { returnValues->addReturnValue(r);                                    };

    void setReturnValueList(ReturnValueList* r)
      { returnValues->clear();
        int n = r->getSize();
        for (int i = 0; i < n; i++)
	    returnValues->addReturnValue(r->getReturnValue(i));
      };

    void clearReturnValueList()   { returnValues->clear();              };
    ReturnValueString* removeReturnValue(int i) 
      { return returnValues->removeReturnValue(i); };
    void removeReturnValue(ReturnValueString* s) 
      { returnValues->removeReturnValue(s);   };

    void addErrorMessage(ErrorMessageString* e)
      { errorMessages->addErrorMessage(e); };

    void setErrorMessageList(ErrorMessageList* e)
      { errorMessages->clear();
        int n = e->getSize();
        for (int i = 0; i < n; i++)
	    errorMessages->addErrorMessage(e->getErrorMessage(i));
      };

    void clearErrorMessageList()  { errorMessages->clear();               };
    ErrorMessageString* removeErrorMessage(int i)
                           { return errorMessages->removeErrorMessage(i); };
    void removeErrorMessage(ErrorMessageString* s) 
                           { errorMessages->removeErrorMessage(s);        };

    String* makeResponseString();

    Boolean sendResponse() 
      { Boolean ret;
        String* dataStr = makeResponseString();
        String* domainStr  = makeMBDomainSpecString(getReceiver());
        MB_sendMessage(dataStr, domainStr);
        if (TMB_ERROR->found())
           ret = FALSE;
        else
           ret = TRUE;
        delete(dataStr);
        delete(domainStr);
        return ret;
      };
};





class Command : public Message {
      MessageType   type;
      int           command;
      ArgumentList* arguments;

      const char* getTypeString();

    public:
      // constructor
      // -----------
      Command() : Message()
        { type = COMMAND;
          addReceiver(MESSAGE_SERVER);
          arguments = new ArgumentList;
        };
      Command(int c) : Message() 
        { type = COMMAND;
          addReceiver(MESSAGE_SERVER);
          command = c;
          arguments = new ArgumentList;
        };
      Command(int c, ArgumentList* a) : Message()
        { type = COMMAND;
          command = c;
          addReceiver(MESSAGE_SERVER);
          arguments = new ArgumentList;

          int n = a->getSize();
          for (int i = 0; i < n; i++)
              arguments->addArgument(new ArgumentString(*(a->getArgument(i))));
        };
      
      // Read Only Routines
      // ------------------
      int            getCommand()      { return command;   };
      ArgumentList*  getArgumentList() { return arguments; };
      MessageType    getType()         { return type;      };

      void printCommand(int c, ostream& outStream = cout);

      void print (ostream& outStream = cout);

      // Write Access Routines
      // ---------------------
      void setArguments(ArgumentList* a)
        { arguments->clear();
          int n = a->getSize();
          for (int i = 0; i < n; i++)
              arguments->addArgument(a->getArgument(i));
        };
      void addArgument(ArgumentString* a)
        { arguments->addArgument(a);
        };
      void setCommand(int c)
        { command = c;
        };

      Boolean sendCommand();
};



//////////////////////////////////////////////////////////////////////////////
//                                                                          //
//                                Queues                                    //
//                                                                          //
//////////////////////////////////////////////////////////////////////////////

class Queue : public GenericList
{
  public:
    // read only routines
    Boolean empty() { return (getSize() == 0); };
    Generic* front(){ return getItem(0);       };

    // write access routines
    void join(Generic* i)  { addItem(i);       };
    void leave()           { removeItem(0);    };
};
 

class RequestQueue : public Queue
{
  public:
    // read only routines
    Boolean empty() { return Queue::empty();           };
    Request* front(){ return (Request*)Queue::front(); };

    // write access routines
    void join(Request* i)  { Queue::join(i);           };
    void leave()           { Queue::leave();           };
};
 

class ResponseQueue : public Queue
{
  public:
    // read only routines
    Boolean empty()  { return Queue::empty();            };
    Response* front(){ return (Response*)Queue::front(); };

    // write access routines
    void join(Response* i)   { Queue::join(i);           };
    void leave()             { Queue::leave();           };
};
 

class MonologueQueue : public Queue
{
  public:
    // read only routines
    Boolean empty()   { return Queue::empty();            };
    Monologue* front(){ return (Monologue*)Queue::front();};

    // write access routines
    void join(Monologue* i)   { Queue::join(i);           };
    void leave()              { Queue::leave();           };
};
 
extern Client *CLIENT;
extern Server *SERVER;
    
extern RequestQueue   *REQUEST_QUEUE;
extern ResponseQueue  *RESPONSE_QUEUE;
extern MonologueQueue *MONOLOGUE_QUEUE;

extern tml_Method *TDS_TELL;
extern tml_Method *TDS_ASK;
extern tml_Method *TDS_COMMAND;

extern ToolClass* TELOS_DATA_SERVER;
extern ToolClass* REFINE;
extern ToolClass* RIGI;
extern ToolClass* MESSAGE_SERVER;

void TMLinit();
NewMessageStatus TMBgetNewMessage();
NewMessageStatus TMBwaitForMessage();

