// $Author: jmartin $
// $Date: 1999-06-11 15:32:59 $
// $Source: /root/cvs/Rigi/rigiedit/cl_avlset.C,v $
// $Locker:  $
// $State: Exp $
//
// ------------------------------------------------------------------------

//
//	cl_avlset.C: simple object-oriented AVL treesets.
//	These data structures permit removal as well as insertion
//	and location. Memory use is approx. halved by not premitting
//	explicit AvlRecords for tree leaves. This complicates the code
//	considerably.
//

#ifndef __STDIO_H
#include <stdio.h>
#endif
#ifndef AVLSET_H
#include "cl_avlset.h"
#endif

enum { EVEN, LEFT, RGHT };
enum { STORE, REMOVE, FIND };
enum { AVLNODE, AVLINFO };

class AvlRecord {

    public:
        char lean;
        char lefttype, rghttype;
        char dummy;

        void *info;
        AvlRecord *left;
        AvlRecord *rght;
};

// We can simply use a record to hold necessary iteration info

class AvlPos : public AvlRecord {};


// One static recursive routine does all the work. This structure
// is used for in/out information whilst this recursion be happening.

typedef struct
{
    void*	info;		// info to store/return
    void*	key;		// key for store/find/remove
    AvlFunc	func;		// set ordering relation
    int		flag;		// type of operation
    int		type;		// the type of AVL set (duplicates allowed?)
    AvlRecord*	remove;		// node to be replaced with a leaf

} Parm;

static Parm sparm;	// Use this parm first to avoid allocation
static Parm *SPP = NULL;

// macro-ified in case it is necessary to overload this field later
#define	set_lean(x,l)	(x)->lean = (l)
#define	get_lean(x)	((x)->lean)

// local memory allocation

static AvlRecord*  NewRecord( Info* );
static Parm*       NewParm();
static void        FreeParm( Parm* );
static int         FreeRecord( AvlRecord* );
static int         FreeRecords( AvlRecord**,int rtype );

static int dflt_compare( void *k, Info *n ) { return (int)k-(int)n; }

static int inorderflag = 0;   // HACK  bad

static int Rotate( AvlRecord**, int ); // Return zero iff height changes


// Visit() is the main recursive tree traversal routine.
// If type == AVLINFO, it means that *tree points directly to info,
// not to a normal avl record.

// Return + if subtree gets longer, - if shorter, 2 if parent nodetype
// changes (ie. { -2, -1, 0, 1, 2 })

static int Visit( AvlRecord **tree, Parm *P, int rtype )
{
    int i=EVEN,j;
    static int calls = 0;
    int call = calls++;

    AvlRecord *n = (rtype==AVLINFO) ? NULL  : *tree;
    void* info  = (rtype==AVLINFO) ? (void*)*tree : (n?n->info:NULL);

//if (P->flag==REMOVE)
//printf("    visit %2d: key=%d value=%d  ptype=%d info=%d n=%d lean=%d\n",
//call,P->key,P->info,rtype, info,n, n?n->lean:(-1));

    if ( n && !info) printf("PROBLEM in AvlSet!\n");

    if ( info==NULL ) // is this code actually ever executed?
    {
        // If doing anything other than store ... we didn't find it
        if (P->flag!=STORE) { P->info = NULL; return 0; }

        // Stick new info into appropriate place (!?!)
        *tree = (AvlRecord*)P->info;

        P->info = NULL; // Return value is null, indicating a new set member

        // This subtree just got heavier (but no new record was created)
        return 1;
    }

    if (P->remove && P->flag==REMOVE)
    {
        // We've already found the node to be removed, and have made a move
        // right. Now we make a move left, unless we've got replacement info.

        if (rtype==AVLINFO || !n->left) // found the replacement info
        {
//printf("    remove: found replacement info %d, n=%d\n",info,n);
            (P->remove)->info = info;

            // Whoever points here should now point to ...
            (*tree) = n ? n->rght : NULL;

            if (n) FreeRecord( n );
            return -2;  // tree got shorter; pointing to info now
        }
        j = -1; // look left for a replacement node
    }
    else j = (*(P->func))( P->key, info ); // comparison


    // did we find a matching record
    if (j==0) switch (P->flag)
    {
        case STORE:

            if (P->type==0)  // duplicate comparison values not allowed
            {
                void *t = info;  // swap info
                if (n) n->info = P->info; else *tree=(AvlRecord*)P->info;

                P->info = t;  // Return original info
                return 0;
            }

            // If duplicate comparisons are allowed, compare on info ptrs
            j = dflt_compare( P->info, info );
            if (j==0) return 0;
            break;

        case FIND:

            // What about looking for PARTICULAR info?? (same as REMOVE)
            P->info = info;
            return 0;

        case REMOVE:

//printf("    remove: found potential info to remove, P=%d %d\n",P->info,info);
            j = dflt_compare( P->info, info );
            if (P->type) if (j!=0 && P->info)
                break; // keep looking if "duplicate" members allowed

            P->info = info;

//rintf("    remove: found info to remove %d, n=%d\n",info,n);
            if (!n || !n->rght) // if a leaf, we be done
            {
                *tree = n ? n->left : NULL;
                if (n) FreeRecord( n );
                return -2;
            }
            P->remove = n;
            j = 1;  // Look right for node to replace node to be removed
    }

    // Process "virtual" nodes as a special case. If storing a new value,
    // we need the current node to be real.

    if (rtype==AVLINFO)
    {
        // If doing anything other than store ... we didn't find it
        if (P->flag!=STORE) { P->info = NULL; return 0; }

        *tree = n = NewRecord( info );
        if (j<0) { n->left = (AvlRecord*)P->info; n->lean = LEFT; }
            else { n->rght = (AvlRecord*)P->info; n->lean = RGHT; }

        P->info = NULL;
        return 2;  // This subtree got heavier. Tell parent to change rtype.
    }

    // Recur on a subtree. Ensure we know what we're pointing at, as we
    // might have created a new node just underneath.

    if (j<0)
    {
        i = Visit( &(n->left), P, n->lefttype );
        if (i==2) n->lefttype = AVLNODE;
        else if (i==-2) n->lefttype = AVLINFO;

        if (i>0) i = i ? LEFT : EVEN;  // did left subtree get longer?
        else i = i ? RGHT : EVEN;
    }
    if (j>0)
    {
        i = Visit( &(n->rght), P, n->rghttype );
        if (i==2) n->rghttype = AVLNODE;
        else if (i==-2) n->rghttype = AVLINFO;

        if (i>0) i = i ? RGHT : EVEN;  // did rght subtree get longer?
        else i = i ? LEFT : EVEN;
    }

    // Check to see if removal of child caused this node to become superfluous
    if (P->flag==REMOVE) if (!n->left && !n->rght)
    {
        (*tree) = (AvlRecord*)n->info;
        FreeRecord( n );
        return -2;
    }

    if (i==EVEN) return 0;

    if (i!=get_lean(n))
    {
        // this node not a rotation pivot - just set the lean factor
        if (get_lean(n)!=EVEN) set_lean(n,EVEN); else set_lean(n,i);

        // Does this affect the nodes above?
        if (P->flag==STORE) return (get_lean(n)==EVEN) ? 0 : 1;
        if (P->flag==REMOVE) return (get_lean(n)==EVEN) ? -1 : 0;

        fprintf(stderr,"error: AVL, unexpected state\n");
        return 0;
    }

    // A rotation may be required. 3 ptrs to change for single, 5 for double.
    // Height may not change if removing, indicated by zero return value.

    i = Rotate( tree, i );

    if (P->flag==STORE) return 0;
    if (P->flag==REMOVE) return i ? 0 : -1;

    fprintf(stderr,"error: AVL, unexpected state after rotation\n");
    return 0;
}

AvlSet::AvlSet()
{
    size = 0;
    contents = NULL;
}

AvlSet::AvlSet( int t )
{
    size = t&1;
    contents = NULL;
}

AvlSet::~AvlSet()
{
    FreeRecords( &contents, Size()==1?AVLINFO:AVLNODE );
}

static int nr_calls=0, fr_calls=0;

int AvlSet::Empty()
{
    int i = FreeRecords( &contents, Size()==1?AVLINFO:AVLNODE );
    size &= 1;
    contents = NULL;
    return i;
}

void *AvlSet::StoreInfo( void *key, void *info, AvlFunc f )
{
    // See if screwing up a traversal
    if ( inorderflag )
    {
        fprintf(stderr,
        "error: AvlSet, attempt to store while iterating\n");
        return NULL;
    }
//printf("StoreInfo: key=%d info=%d\n",key,info);

    // Store this info in the set, return 0 if already there
    Parm *P = NewParm();
    P->info = info; P->key = key;
    P->type = size&1;
    P->func = f ? f : dflt_compare;
    P->flag = STORE;

    Visit( &contents, P, Size()==1?AVLINFO:AVLNODE );

    if (!P->info) size+=(1<<8);
    info = P->info;
    FreeParm(P);
    return info;
}

void *AvlSet::RemoveInfo( void *key, void *info, AvlFunc f )
{
    // See if screwing up a traversal
    if ( inorderflag )
    {
        fprintf(stderr,"error: AvlSet, attempt remove while iterating\n");
        return NULL;
    }

//printf("removing, key is %d, info is %d, f is %d\n",key,info,f);
    // Find and remove this info, return 0 if not there
    Parm *P = NewParm();
    P->remove = NULL;
    P->info = info; P->key = key;
    P->type = size&1;
    P->func = f ? f : dflt_compare;
    P->flag = REMOVE;

    Visit( &contents, P, Size()==1?AVLINFO:AVLNODE );

    if (P->info) size-=(1<<8);
    info = P->info;
    FreeParm(P);
    return info;
}

void* AvlSet::FindInfo( void *key, AvlFunc f )
{
    // Return first info found with this key
    Parm *P = NewParm();
    P->info = NULL; P->key = key;
    P->func = f ? f : dflt_compare;
    P->flag = FIND;

    Visit( &contents, P, Size()==1?AVLINFO:AVLNODE );

    void *info = P->info;
    FreeParm(P);
    return info;
}

static int maxlevel;
static int numset;
static int numslots;
static int unbalanced;
static int badleanfactor;
static int leaves;
static int halfleaves;

static int PrintSet( AvlRecord **tree, int level, void p(void*), int pt )
{
    // Return the maximum depth (real nodes only) reached from here
    int L=0,R=0;

    AvlRecord *n = (pt==AVLINFO) ? NULL : *tree;
    void *info = (pt==AVLINFO) ? *tree : (n?n->info:NULL);
    if (info==NULL) return 0;

    numslots++;
    numset++;
    if (level>maxlevel) maxlevel = level;

    if (n) L = PrintSet( &(n->left), level+1, p, n->lefttype );

    if (p)
    {
        printf("%2d%c %7d:  ",level,
            n?(get_lean(n)==EVEN?'E':(get_lean(n)==LEFT?'L':'R')):' ',
            info);
        p(info); putchar('\n');
    }

    if (n) R = PrintSet( &(n->rght), level+1, p, n->rghttype );

    if ((R-L)>1 || (L-R)>1) if (level>unbalanced)
    {
        unbalanced = level;
        printf("error: AVL imbalance at level %d. ",level);
        if(p)p(info);putchar('\n');
    }
    if (n && (R-L)==1 && get_lean(n)!=RGHT) badleanfactor = level;
    if (n && (L-R)==1 && get_lean(n)!=LEFT) badleanfactor = level;
    if (n && (R-L)==0 && get_lean(n)!=EVEN) badleanfactor = level;

    if (!n) leaves++;
    if (R!=L && (L==0||R==0)) halfleaves++;
    return L>R ? L+1 : R+1;
}

void AvlSet::Count() { Print(NULL); }
void AvlSet::Print( void (*p)(void*) )
{
    numset = numslots = maxlevel = leaves = halfleaves = 0;
    badleanfactor = unbalanced = -1;

    int d = PrintSet( &contents,0,p, Size()==1?AVLINFO:AVLNODE );

    if (unbalanced>=0)
        fprintf(stderr,"error: AvlSet, unbalanced set detected!\n");

    if (badleanfactor>=0)
        fprintf(stderr,"error: AvlSet, bad lean factor at level %d\n",
        badleanfactor);

    printf("total=%d slots=%d maxdepth=%d leaves=%d halfleaves=%d\n",
        numset,numslots,maxlevel,leaves,halfleaves );
}

Info *AvlSet::First()
{
    AvlRecord *R = contents;
    if (Size()<=1) return (Info*)R;

    int recordtype = AVLNODE;
    for ( ; recordtype==AVLNODE && R->left; R=R->left )
        recordtype = R->lefttype;

    return (recordtype==AVLINFO) ? (Info*)R : R->info;
}

Info *AvlSet::Last()
{
    AvlRecord *R = contents;
    if (Size()<=1) return (Info*)R;

    int recordtype = AVLNODE;
    for ( ; recordtype==AVLNODE && R->rght; R=R->rght )
        recordtype = R->rghttype;

    return (recordtype==AVLINFO) ? (Info*)R : R->info;
}


// Algorithm for First()/Next(): use a chain of records to keep track of
// a path in the tree, providing the reverse links which the tree itself
// does not provide. Records in the chain point to both real and "virtual"
// nodes.

Info *AvlSet::First( AvlPos*& pos )
{
    AvlRecord *R, *P;
    pos = NULL;
    if (!contents) return NULL;

    // Go left as far as possible
    int rtype = (Size()==1) ? AVLINFO : AVLNODE;
    for ( R=contents; R; R=R->left )
    {
        // Put a new stack entry on the list
        P = NewRecord(NULL);
        P->left = R;
        P->lefttype = rtype;
        P->rght = pos;
        pos = (AvlPos*)P;

        if (rtype==AVLINFO) break;
        rtype = R->lefttype;
    }
    R = pos->left;
    return (pos->lefttype==AVLINFO) ? (Info*)R : R->info;
}

Info *AvlSet::Next( AvlPos*& pos )
{
    // Note: no use of instance variables here, just "pos"
    if (!pos) return NULL;

    AvlRecord *R=pos->left, *P=NULL;

    // Go right once in the tree, then left as far as possible
    if (pos->lefttype==AVLNODE && R->rght)
    {
        int rtype = R->rghttype;
        for ( R=R->rght; R; R=R->left )
        {
            // Put a new stack entry on the list
            P = NewRecord(NULL);
            P->left = R;
            P->lefttype = rtype;
            P->rght = pos;
            pos = (AvlPos*)P;

            if (rtype==AVLINFO) break;
            rtype = R->lefttype;
        }
        R = pos->left;
        return (pos->lefttype==AVLINFO) ? (Info*)R : R->info;
    }
    else
    {
        // Pop the stack until we make a backwards move right
        do
        {
            P = pos;
            pos = (AvlPos*)P->rght;
            R = P->left;
            FreeRecord( P );

        } while (pos && (pos->left)->rght==R);

        if (pos) return (pos->left)->info;
        return NULL;
    }
}

Info* AvlSet::LastInfo( AvlPos*& P )
{
    if (!P) return NULL;
    return (P->lefttype==AVLINFO) ? (Info*)P->left : P->left->info;
}

void AvlSet::Delete( AvlPos*& P)
{
  while (P) {
    AvlPos* tmp = P;
    P = (AvlPos*)P->rght;
    FreeRecord(tmp);
  }
}

Info* AvlSet::First( AvlPos*& pos, void *key, AvlFunc f )
{
    AvlRecord *R=contents, *P;
    Info *info;
    AvlFunc compare;
    compare = f ? f : dflt_compare;

    pos = NULL;
    if (!R) return NULL;

    // Go down tree as far as possible, making sure we stay LEFT of the
    // node we want.

    int rtype = (Size()==1) ? AVLINFO : AVLNODE;
    while (R)
    {
        // Put a new stack entry on the list
        P = NewRecord(NULL);
        P->left = R;
        P->lefttype = rtype;
        P->rght = pos;
        pos = (AvlPos*)P;

        info = (rtype==AVLINFO) ? (Info*)R : R->info;
        if (rtype==AVLINFO) break;

        if ((*compare)(key,info)>0)
        {
            rtype = R->rghttype;
            R = R->rght;
        }
        else
        {
            rtype = R->lefttype;
            R = R->left;
        }
    }

    // Now, get first info that is greater or equal to the key.

    while (info && (*compare)(key,info)>0) info = Next(pos);
    return info;
}

static void *Inorder( AvlRecord *R, int (*f)(void*), int rtype )
{
    void *A, *I = (rtype==AVLINFO) ? (void*)R : (R?R->info:NULL);
    if ( I==NULL ) return NULL;

    if (R && rtype==AVLNODE)
    if ( A=Inorder( R->left,f,R->lefttype ) ) return A;

    if ( (*f)(I) ) return I;

    if (R && rtype==AVLNODE)
    if ( A=Inorder( R->rght,f,R->rghttype ) ) return A;

    return NULL;
}

void* AvlSet::Iterate( int (*f)(void*) )
{
    inorderflag = 1;
    void *r = Inorder( contents, f, Size()==1?AVLINFO:AVLNODE );
    inorderflag = 0;
    return r;
}

////////////////////////////////////////////////////////////////

// MEMORY ALLOCATION

static AvlRecord *rlist = NULL;    // List of free nodes
static int mem = 0;             // Number of bytes actually allocated

int GetAvlObjMem() { return mem; }

static AvlRecord *NewRecord( Info *I )
{
    static int maxindex = 64;
    static int index = maxindex;
    static const int maxblock = 512;
    static AvlRecord *array = NULL;

    AvlRecord *r;

    nr_calls++;
    if (rlist) {
        r = rlist; rlist = r->rght;
    } else if (index<maxindex) {
        r = array + index++;
    } else {
        if ( maxindex < maxblock ) maxindex<<=1;
        array = new AvlRecord[maxindex];
        mem += maxindex*sizeof(AvlRecord);
        index = 0;
        r = array + index++;
    }
    r->info = I;
    r->left = r->rght = NULL;
    r->lean = EVEN;
    r->lefttype = r->rghttype = AVLINFO;
    return r;
}

static Parm *NewParm()
{
    if (SPP==NULL) return SPP=&sparm;

    fprintf(stderr,"warning: AvlSet, interleaved set traversals (OK)\n");
    Parm *p = new Parm;
    return p;
}

static void FreeParm( Parm *P )
{
    if (SPP==P) SPP = NULL;
    else delete P;
}

static int FreeRecord( AvlRecord *r )
{
    fr_calls++;

    // add the AVL record to the free list
    r->rght = rlist; rlist = r;
    return 1;
}

static int FreeRecords( AvlRecord **tree, int rtype )
{
    int r=0;
    AvlRecord *n = (rtype==AVLINFO) ? NULL : *tree;
    if (!n) return 0;

    r += FreeRecords( &(n->left), n->lefttype );
    r += FreeRecords( &(n->rght), n->rghttype );

    r += FreeRecord( n );
    return r;
}


//////////////////////////////////////////////////////////////////////////////
/*
    This is the difficult part of this AVL tree data structure. The tree must
    be "rotated" iff the lean factor of the current node is not EVEN, and
    equal to the lean factor inherited from action happening below the current
    node (this action has already taken place). Note that it is possible for
    no change in overall height to occur, even after a rotation has taken
    place, when a balanced tree has a node removed.
    Returns zero iff height changes (most common case).
*/
//////////////////////////////////////////////////////////////////////////////

static int Rotate( AvlRecord **tree, int lean )  // lean is either LEFT or RGHT
{
    AvlRecord *n = *tree; int r=0;
    if (!n) { fprintf(stderr,"error: AVL::Rotate\n"); return 0; }

    if (lean==RGHT && get_lean(n->rght)!=LEFT)
    {
//printf("    rotating left simple\n");
        r = get_lean(n->rght)==EVEN;
        set_lean(n,r?RGHT:EVEN);
        set_lean(n->rght,r?LEFT:EVEN);

        *tree = (n->rght);
        n->rght = (n->rght)->left;
        (*tree)->left = n;

        n->rghttype = (*tree)->lefttype;
        (*tree)->lefttype = AVLNODE;
    }
    else if (lean==LEFT && get_lean(n->left)!=RGHT)
    {
//printf("    rotating rght simple\n");
        r = get_lean(n->left)==EVEN;
        set_lean(n,r?LEFT:EVEN);
        set_lean(n->left,r?RGHT:EVEN);

        *tree = (n->left);
        n->left = (n->left)->rght;
        (*tree)->rght = n;

        n->lefttype = (*tree)->rghttype;
        (*tree)->rghttype = AVLNODE;
    }
    else if (lean==RGHT && get_lean(n->rght)==LEFT)
    {
        // degenerate case needs to make a new record
        int d = (n->rght)->lefttype==AVLINFO;
//printf("    rotating left double %s\n",d?"INFO":"NODE");
        (*tree) = d ? NewRecord( (Info*)((n->rght)->left) ) : (n->rght)->left;

        (n->rght)->left = (*tree)->rght;
        (*tree)->rght = n->rght;
        n->rght = (*tree)->left;
        (*tree)->left = n;

        set_lean(n,get_lean(*tree)!=RGHT ? EVEN : LEFT );
        set_lean((*tree)->rght, get_lean(*tree)!=LEFT ? EVEN : RGHT );
        set_lean(*tree,EVEN);

        n->rghttype = (*tree)->lefttype;
        ((*tree)->rght)->lefttype = (*tree)->rghttype;
        (*tree)->lefttype = (*tree)->rghttype = AVLNODE;
    }
    else if (lean==LEFT && get_lean(n->left)==RGHT)
    {
        // degenerate case needs to make a new record
        int d = (n->left)->rghttype==AVLINFO;
//printf("    rotating rght double %s\n",d?"INFO":"NODE");
        (*tree) = d ? NewRecord( (Info*)((n->left)->rght) ) : (n->left)->rght;

        (n->left)->rght = (*tree)->left;
        (*tree)->left = n->left;
        n->left = (*tree)->rght;
        (*tree)->rght = n;

        set_lean(n,get_lean(*tree)!=LEFT ? EVEN : RGHT);
        set_lean((*tree)->left,get_lean(*tree)!=RGHT ? EVEN : LEFT);
        set_lean(*tree,EVEN);

        n->lefttype = (*tree)->rghttype;
        ((*tree)->left)->rghttype = (*tree)->lefttype;
        (*tree)->rghttype = (*tree)->lefttype = AVLNODE;
    }
    else
    {
        fprintf(stderr,"error: inconsistent state in AVL tree-set!\n");
        fprintf(stderr,"error: lean=%d n->lean=%d n=%d n->left=%d n->rght=%d\n",
            lean,get_lean(n),n,n->left,n->rght);
    }


    // One or both children may have become leaves after a simple rotation.
    // If this is so, must free their records and update parent's ptr info.

    AvlRecord *m; n = *tree;
    if (n->lefttype==AVLNODE)
    if (n->left->left==NULL && n->left->rght==NULL)
    {
//printf("    degenerate left case\n");
        m = n->left;
        n->left = (AvlRecord*)m->info; n->lefttype = AVLINFO;
        FreeRecord( m );
    }
    if (n->rghttype==AVLNODE)
    if (n->rght->left==NULL && n->rght->rght==NULL)
    {
//printf("    degenerate rght case\n");
        m = n->rght;
        n->rght = (AvlRecord*)m->info; n->rghttype = AVLINFO;
        FreeRecord( m );
    }
    return r;
}
