// $Author: jmartin $
// $Date: 2000-01-12 01:14:01 $
// $Source: /root/cvs/Rigi/rigiedit/cl_domain.C,v $
// $Locker:  $
// $State: Exp $
//
// ------------------------------------------------------------------------
#ifndef __STRING_H
#include <string.h>
#endif
#include <ctype.h>		// for isdigit()


#ifndef GLOBAL_H
#include "global.h"             // for ANY_ARC and ediotr
#endif
#ifndef GRAPHARC_H
#include "cl_grapharc.h"
#endif
#ifndef DOMAIN_H
#include "cl_domain.h"          // Definition for RigiDomain class
#endif
#ifndef NUMARRAYH
#include "adt/numarray.h"
#endif
#ifndef CHARARRAYH
#include "adt/chararray.h"
#endif
#ifndef ATTRIBUTE_H
#include "cl_attribute.h"
#endif
#ifndef GRAPHNODE_H
#include "cl_graphnode.h"
#endif

#ifndef RIGISTATE_H
#include "cl_rigistate.h"		// Global instance of editor's state
#endif
#include "environ.h"            // Environment function calls
#ifndef RIGIERR_H
#include "rigierr.h"
#endif
#ifndef GRAPHUTIL_H
#include "graphutil.h"          // Graph utility routines (file IO)
#endif
#ifndef RIGIUTIL_H
#include "rigiutil.h"           // Editor utility routines
#endif
#ifndef RCL_COMMANDS_H
#include "rcl_commands.h"       // For RCL interface stuff
#endif
#ifndef QUIT_H
#include "quit.h"
#endif


static const int buffsize = 256;
static char buff[buffsize];

//------------------------------------------------------------------------
// Domain Constructor and Destructors
//------------------------------------------------------------------------

RigiDomain::RigiDomain( char *domain )
{
    char *dmn = domain;

    // Empty out the various lists and sets associated with the class

    nodenames.Empty();
    iconfiles.Empty();
    arcnames.Empty();

    nodeset = new AttributeSet();
    nodeattrs = new AttributeSet();

    arcset = new AttributeSet();
    arcattrs = new AttributeSet();

    srctypes.Empty();
    dsttypes.Empty();

    // Initialize the number of node and arc types.

    num_nodetypes = 0;
    num_arctypes = 0;

    // If no model specified, check for a default domain model. If the
    // environment is not up to date, initialize it.

    if (!dmn || !*dmn) dmn = EnvGet( "RIGIDOMAIN" );
    if (!dmn) dmn = "";

    EnvSet( "RIGIDOMAIN", dmn );

    // Initialize the domain color info

    colors = NULL;

    // Read in the domain info from the appropriate external files

    ReadNodeInfo();	// Read Riginode info (before arc info!)
    ReadArcInfo();	// Read Rigiarc info
    ReadAttrInfo();	// Read permissible attributes of nodes and arcs
    ReadColorInfo();	// Read color info for the domain
    ReadRCLInfo();	// Finally, interpret domain-specific RCL code
}

RigiDomain::~RigiDomain()
{
    delete nodeset;
    delete nodeattrs;
    delete arcset;
    delete arcattrs;
}

//------------------------------------------------------------------------
// General domain methods
//------------------------------------------------------------------------

char* RigiDomain::Name() const
{
    return domain.string();
}

//------------------------------------------------------------------------
// Node type access methods.
//------------------------------------------------------------------------

int RigiDomain::NumNodeTypes() const
{
    return num_nodetypes;
}

char* RigiDomain::NodeNameFromType( int n )
{
    return ( n<0 || (unsigned)n>=nodenames.Size()) ?
        NULL : nodenames.At(n);
}

char* RigiDomain::IconFileName( int i )
{
    return ( i<0 || (unsigned)i>=iconfiles.Size()) ?
        NULL : iconfiles.At(i);
}

int RigiDomain::NodeTypeFromName( char *name )
{
    int t = -1;
    if (name && isdigit(*name)) t = atoi(name);
    else t = nodeset->GetIntAttr( name );

    if (t>=0 && t<num_nodetypes) return t;
    return -1;
}

//------------------------------------------------------------------------
// Arc type access methods.
//------------------------------------------------------------------------

int RigiDomain::NumArcTypes() const
{
    return num_arctypes;
}

int RigiDomain::ArcSrcType( int arctype )
{
    return (arctype<0 || (unsigned)arctype>=srctypes.size()) ?
        -1 : srctypes.At(arctype);
}

int RigiDomain::ArcDstType( int arctype )
{
    return (arctype<0 || (unsigned)arctype>=dsttypes.size()) ?
        -1 : dsttypes.At(arctype);
}

char* RigiDomain::ArcNameFromType( int type )
{
    if (type==ANY_ARC) return "any";
    if (type < 0 || (unsigned) type >= arcnames.Size()) return NULL;
    return arcnames.At(type);
}

int RigiDomain::ArcTypeFromName( char *name )
{
    if (!strncmp(name,"any",3)) return ANY_ARC;	// HACK!

    if (name && isdigit(*name)) return atoi(name);
    int r = arcset->GetIntAttr( name );

    return r >= 0? r : -1;
}

void RigiDomain::SetArcColor( int i, RigiColor *col )
{
    int n = NumArcTypes();
    if (i<0 || i>=n) {
        mayday("Arc index %d is out of valid range (0 - %d)", i, n-1);
        return;
    }
    n = NumNodeTypes() + i;
    colors[n].red = col->red;
    colors[n].green = col->green;
    colors[n].blue = col->blue;
}

int RigiDomain::GetArcColor( int i, RigiColor* col )
{
    int n = NumArcTypes();
    if (i<0 || i>=n) {
        mayday("Arc index %d is out of valid range (0 - %d)", i, n-1);
        return 0;
    }
    n = i + NumNodeTypes();
    col->red = colors[n].red;
    col->green = colors[n].green;
    col->blue = colors[n].blue;
    return 1;
}

void RigiDomain::SetNodeColor( int i, RigiColor *col )
{
    int n = NumNodeTypes();
    if (i<0 || i>=n) {
        mayday("Node index %d is out of valid range (0 - %d)", i, n-1);
        return;
    }
    colors[i].red = col->red;
    colors[i].green = col->green;
    colors[i].blue = col->blue;
}

int RigiDomain::GetNodeColor( int i, RigiColor* col )
{
    int n = NumNodeTypes();
    if (i<0 || i>=n) {
        mayday("Node index %d is out of valid range (0 - %d)", i, n-1);
        return 0;
    }
    col->red = colors[i].red;
    col->green = colors[i].green;
    col->blue = colors[i].blue;
    return 1;
}


//------------------------------------------------------------------------
// functions for reading domain files
//------------------------------------------------------------------------

int RigiDomain::ReadArcInfo()
{
    FILE *fp = NULL;
    int newsyntaxflag = 0;
    int narctypes;

    CharType filebuff;
    filebuff.Assign( EnvDomainDir() );
    filebuff.Append( "/Rigiarc" );
    fp = fopen( filebuff.string(), "r" );

    if ( !fp )
    {
        narctypes = 8;
        warn( "couldn't open arc file %s, using %d arc types",
            filebuff.string(),narctypes );
    }
    else
    {
        narctypes = 0;
        RutilFscanl( fp,buff,buffsize );
        if (isdigit(buff[0])) narctypes = atoi(buff);
            else newsyntaxflag = 1;
    }

    // Read in the arc tuples [name type].
    // The values for the arctype should match those returned from the
    // database. In this way the user can change the representations of the
    // info in the database. Currently, the arc type values are ignored.

    char arctypename[64], buff2[64], buff3[64], buff4[64];

    char *srctypename, *dsttypename;
    int arctype=0, srctype=0, dsttype=0, nrd=0;

    int i;
    for ( i=0; 0==narctypes || i<narctypes; i++ )
    {
        if (fp==NULL)
        {
            arctype = i;
            sprintf( arctypename, "arctype%d", arctype );
        }
        else
        {
            if (newsyntaxflag) newsyntaxflag = 0;
                else if (!RutilFscanl( fp, buff, buffsize )) break;

            nrd = sscanf( buff,"%s %s %s %s", arctypename, buff2,buff3,buff4 );

            srctypename = NULL; dsttypename = NULL;
            if ( nrd==3 )
            {
                srctypename = buff2;
                dsttypename = buff3;
            }
            else if ( nrd==4 )
            {
                srctypename = buff3;
                dsttypename = buff4;
            }
            else
            {
                srctypename = NULL;
                dsttypename = NULL;
            }

            // Get the nodetypes which this relation is for (if any declared)
            srctype = srctypename ? NodeTypeFromName( srctypename ) : 0;
            dsttype = dsttypename ? NodeTypeFromName( dsttypename ) : 0;

            srctypes.Create( srctype );
            dsttypes.Create( dsttype );

            if ( nrd<1 || nrd>4 )
            {
                mayday("corrupt file %s, %d chars read",filebuff.string(),nrd);
                fp = NULL; i--; continue;
            }
        }
        NewArcType( arctypename );
    }
    if ( nrd==2 || nrd==4 )
        warn( "archaic syntax in file %s",filebuff.string() );

    if ( narctypes && i!=narctypes )
        warn( "%d arctypes read, but %d declared",i,narctypes );

    if ( i!=num_arctypes )
        mayday("arctype count problem in ReadArcInfo");


    // MJW HACKS: must have composite and level arc types

    int lt = ArcTypeFromName("composite");
    if ( lt == -1 )
    {
        lt = NewArcType( "composite" );
        warn( "no arctype \"composite\" declared, using type %d",lt );
    }
    editor->SetCompositeArc( lt );

    lt = ArcTypeFromName("level");
    if ( lt == -1 )
    {
        char *s = ArcNameFromType(0);
        warn( "no arctype \"level\" declared, using type \"%s\"",s );
        lt = 0;
    }
    editor->SetLevelArc( lt );

    if (fp) fclose(fp);
    return 1;
}

int RigiDomain::ReadAttrInfo()
{
    int i;
    FILE *fp = NULL;
    CharType filebuff;

    filebuff.Assign( EnvDomainDir() );
    filebuff.Append( "/Rigiattr" );
    fp = fopen( filebuff.string(), "r" );
    if ( !fp )
    {
        //warn( "couldn't open Rigiattr; no attribute categories declared" );
        return 0;
    }

    for ( ;; )
    {
        if ( !RutilFscans( fp, buff, buffsize ) ) break;
        if ( strcmp(buff,"attr") )
        {
            // syntactic variation: assume this is an Object attribute
                arcattrs->SetAttr( buff,1 );
            nodeattrs->SetAttr( buff,1 );
            continue;
        }

        if ( !RutilFscans( fp, buff, buffsize ) )
        {
            mayday("missing object class in Rigiattr file"); return 0;
        }
        else if (!strcmp(buff,"Object"))	i=1;
        else if (!strcmp(buff,"Node"))		i=2;
        else if (!strcmp(buff,"Arc"))		i=3;
        else if (ArcTypeFromName(buff)!=-1)	i=4;
        else if (NodeTypeFromName(buff)!=-1)	i=5;
        else					i=6;

        if ( !RutilFscans( fp, buff, buffsize ) )
        {
            mayday("missing attribute name in Rigiattr file"); return 0;
        }
        else switch (i)
        {
            case 0:
                break;
            case 1:
                arcattrs->SetAttr( buff,1 );
                nodeattrs->SetAttr( buff,1 );
                break;
            case 2:
                nodeattrs->SetAttr( buff,1 );
                break;
            case 3:
                arcattrs->SetAttr( buff,1 );
                break;
            case 4:
                nodeattrs->SetAttr( buff,1 );
                break;
            case 5:
                arcattrs->SetAttr( buff,1 );
                break;
            case 6:
                mayday("unknown class \"%s\" in Rigiattr file",buff);
                break;
            default:
                break;
        }
    }
    fclose(fp);
    return 1;
}

int RigiDomain::ReadNodeInfo()
{
    char typeName[MAXPATHLEN+1];
    char iconfile[MAXPATHLEN+1];
    FILE *fp=NULL;
    int newsyntaxflag = 0;
    int nodetype, i, nrd=0;

    CharType filebuff;
    filebuff.Assign( EnvDomainDir() );
    filebuff.Append( "/Riginode" );
    fp = fopen( filebuff.string(), "r" );

    if ( !fp )
    {
        i = 4;
        warn( "couldn't open node file %s, using %d node types",
            filebuff.string(), i );
    }
    else
    {
        i = 0;
        RutilFscanl( fp, buff, buffsize );
        if (isdigit(buff[0])) i = atoi( buff );
                else newsyntaxflag = 1;
    }

    // read in nodetype names and icon file names
    int n;
    for ( n=0; i==0 || n<i; n++ )
    {
        if ( !fp )
        {
            sprintf( typeName,"nodetype%d",n );
            nodetype = n;
        }
        else
        {
            if (newsyntaxflag) newsyntaxflag = 0;
                else if (!RutilFscanl( fp, buff, buffsize )) break;

            nrd = sscanf( buff, "%s %s %d", typeName,iconfile,&nodetype );
            if ( nrd !=1 && nrd!=2 && nrd!=3 )
            {
                mayday( "corrupt node file %s",filebuff.string() );
                quit(0);
            }
            if ( nrd==1 || nrd==2) nodetype = -1;
        }
        NewNodeType( typeName );

        // save file name for this icon

        if (nrd >= 2) {
            char *icondir = EnvGet("ICONDIR");
            if (icondir != NULL) 
                sprintf( typeName, "%s/%s", icondir, iconfile );
            else
                *typeName = 0;
        } else {
            *typeName = 0;
        }
        iconfiles.Create(typeName);
    }
    if ( nrd==3 )
        warn( "archaic syntax in file %s",filebuff.string() );

    if ( n!=num_nodetypes)
        mayday("problem with nodetype count %d vs %d",n,num_nodetypes );

    if ( i && i!=n )
        warn("%d nodetypes declared, but %d read",i,n );

    // We need to know what to use for "collapse" nodes

    int ct = NodeTypeFromName("Collapse");
    if ( ct == -1 )
    {
        ct = NewNodeType( "Collapse" );
        warn( "no nodetype \"Collapse\" was declared, using type %d",ct );
    }
    editor->SetCollapseNode(ct);

    if (fp) fclose(fp);

    return 1;
}

void RigiDomain::ReadRCLInfo()
{
    CharType filebuff;
    filebuff.Assign( EnvDomainDir() );
    filebuff.Append( "/Rigircl" );
    FILE *fp = fopen( filebuff.string(), "r" );
    if (fp)
    {
        fclose(fp);
        CharType cmd;
        cmd.Assign( "source " );
        cmd.Append( filebuff.string() );
        cmd.Append( "\n" );
        Mess("loading %s",filebuff.string() );
        exec_command( cmd.string() );
    }
}

int RigiDomain::ReadColorInfo()
{
    FILE *col_fp;
        CharType filebuff;
    RigiColor col;
    int i;
    int c1, c2, c3;

        // Delete and reinialize the color table

        if (colors) delete[] colors;
        ncolors = NumNodeTypes() + NumArcTypes();
    colors = new RigiColor[ncolors];

    // Open the color file

    filebuff.Assign( EnvDomainDir() );
    filebuff.Append( "/Rigicolor" );
        col_fp = fopen( filebuff.string(), "r" );
    if ( !col_fp )
        warn( "can't read color file %s, using defaults",filebuff.string() );


        // If the color file could not be opened, fill the color table with
    // a default color, red in this case.

    if ( col_fp == NULL )
    { 
        col.red = 0xffff; col.green = col.blue = 0;
        for ( i=0; i<NumNodeTypes(); i++ )
            SetNodeColor(i,&col);
        for ( i=0; i<NumArcTypes(); i++ )
            SetArcColor( i,&col );
    } else {
        for ( i=0; i<NumNodeTypes(); i++ )
        {
            RutilFscanl( col_fp, buff, buffsize );
            if ( sscanf( buff, "%d%d%d", &c1, &c2, &c3 ) != 3 )
                mayday( "problem in color file %s", filebuff.string() );
            col.red = c1;
            col.green = c2;
            col.blue = c3;
            SetNodeColor( i, &col );
        }
        for ( i=0; i<NumArcTypes(); i++ )
        {
            RutilFscanl( col_fp, buff, buffsize );
            if ( sscanf( buff, "%d%d%d", &c1, &c2, &c3 ) != 3 )
                mayday("problem in color file %s (arcs)", filebuff.string() );
            col.red = c1;
            col.green = c2;
            col.blue = c3;
            SetArcColor( i, &col );
        }
    }
    fclose( col_fp );

    return 1;
}

int RigiDomain::NewArcType( char *name )
{
    if (name && isdigit(*name))
    {
        mayday("must use an alpha id for arctype names");
        return -1;
    }

    // insert record into the AVL set
    if (arcset->SetAttr( name, arcnames.Size() ))
        mayday("arctype name \"%s\" repeated",name );

    arcnames.Create( name );
    num_arctypes++;
    return arcnames.Size()-1;
}

int RigiDomain::ChangeArcTypeName( char *oldname, char *name )
{
    int type = ArcTypeFromName( oldname );
    if ( type == -1 )
    {
        Mess( "can't change name of unknown arctype \"%s\"",oldname );
        return -1;
    }

    arcset->RemoveAttr( oldname );
    if (strcmp( ArcNameFromType(type), oldname ))
    {
        mayday("problem in ChangeArcTypeName: %d %s %s",type,oldname,name);
        return -1;
    }

    if (arcset->SetAttr(name,type))
    {
        mayday("repeat arctype, name %s for type %d",name,type );
        return -1;
    }

    arcnames.Create( type, name );
    return type;
}

int RigiDomain::NewNodeType( char *name )
{
    if (!name || isdigit(*name))
    {
        mayday("NewNodeType, nodetype names must be alphabetic");
        return -1;
    }
    int type = nodenames.Size();

    if (nodeset->SetAttr(name,type))
        mayday("repeat nodetype name \"%s\", for type %d",name,type );

    nodenames.Create( name );
    num_nodetypes++;
    return nodenames.Size()-1;
}

int RigiDomain::ChangeNodeTypeName( char *oldname, char *newname )
{
    int type = NodeTypeFromName( oldname );
    if ( type == -1 )
    {
        Mess( "can't change name of unknown nodetype \"%s\"",oldname );
        return -1;
    }
    nodeset->RemoveAttr( oldname );

    if (nodeset->SetAttr( newname, type ))
    {
        mayday("repeat nodetype, name %s for type %d", newname, type );
        return -1;
    }
    nodenames.Create( type, newname );
    return type;
}

//------------------------------------------------------------------------
// Node attribute methods
//------------------------------------------------------------------------

int RigiDomain::IsNodeAttr( int nodetype, char *attr )
{
    if (!strcmp(attr,"type")) return 1;
    return NULL!=nodeattrs->GetAttr(attr);
}

int RigiDomain::NumNodeAttrs() { return nodeattrs->Size(); }

const char *RigiDomain::FirstNodeAttr(AvlPos*&P) { return nodeattrs->First(P); }

const char *RigiDomain::NextNodeAttr(AvlPos*&P) { return nodeattrs->Next(P); }

void RigiDomain::SetNodeAttr( GraphNode *n, char *name, char *value )
{
    // Usually, we just use the GraphNode method for storing the attr.
    // However, there are a number of special cases ....

    if ( !n || !name || !value ) {
        mayday("problem in WindowSetNodeAttr()");
        return;
    }

    if (!strcmp( name, "type" )) {
        int nodetype = NodeTypeFromName( value );
        if ( nodetype == -1 ) {
            mayday("nodetype \"%s\" undefined",value);
            return;
        }
        n->SetType( nodetype );
    } else if (!strcmp( name, "id" )) {
        int id = atoi( value );
        if ( id==0 ) warn("id of 0 for node \"%s\"",n->GetLabel());
        GutilChangeNodeId( n, id );
    } else if (!strcmp( name,"name" )) {
        GutilChangeNodeName( n, value );
    } else {
        n->SetAttr( name, value );
    }
}

char* RigiDomain::GetNodeAttr( GraphNode *n, char *name )
{
    if ( !n || !name ) mayday("problem in WindowGetNodeAttr()");

    if (!strcmp(name,"type")) return ArcNameFromType(n->GetType());
    return n->GetAttr(name);
}

//------------------------------------------------------------------------
// Arc attribute methods
//------------------------------------------------------------------------

int RigiDomain::IsArcAttr( int arctype, char *attr )
{
    if (!strcmp(attr,"type")) return 1;
    return NULL!=arcattrs->GetAttr(attr);
}
int RigiDomain::NumArcAttrs() { return arcattrs->Size(); }
const char *RigiDomain::FirstArcAttr(AvlPos*&P) { return arcattrs->First(P); }
const char *RigiDomain::NextArcAttr(AvlPos*&P) { return arcattrs->Next(P); }


RigiBool RigiDomain::operator==( const RigiDomain& d ) const
{
    return (strcmp(Name(), d.Name()) == 0) ? RigiTrue : RigiFalse;
}

void RigiDomain::operator=( const RigiDomain& d )
{
        mayday("RigiDomain::operator= - Not implemented");
}

int RigiDomain::Printout(int) const {
        mayday("RigiDomain::Printout - Not implemented");
        return 0;
}

unsigned int RigiDomain::Hash() const {
    mayday("RigiDomain::Hash - Not implemented");
    return 0;
}

RigiBool RigiDomain::isEqual(void *a) const {
    mayday("RigiDomain::isEqual - Not implemented");
    return RigiFalse;
}

void RigiDomain::SetArcAttr( GraphArc *a, char *name, char *value )
{
    if ( !a || !name || !value ) {
        mayday("problem in WindowSetArcAttr()");
        return;
    }

    if (!strcmp( name, "type" )) {
        int arctype = ArcTypeFromName( value );
        if ( arctype == -1 ) {
            mayday("nodetype \"%s\" undefined",value);
            return;
        }
        a->SetType( arctype );
    } else {
        a->SetAttr( name, value );
    }
}

char* RigiDomain::GetArcAttr( GraphArc *a, char *name )
{
    if ( !a || !name )
        mayday("problem in WindowGetArcAttr()");

    if (!strcmp(name,"type")) return ArcNameFromType(a->GetType());
    return a->GetAttr(name);
}


