//
// File:    graphnode.h
// Author:  Brian Corrie
// Date:    January 17, 1991
// Project: Rigi V4.0
//

// $Author: jmcdanie $
// $Date: 1996-05-24 00:35:10 $
// $Source: /root/cvs/Rigi/rigiedit/cl_graphnode.h,v $
// $Locker:  $
// $State: Exp $

//------------------------------------------------------------------------

#ifndef GRAPHNODE_H
#define GRAPHNODE_H

#ifndef OBJECTH
#include "adt/object.h"        // Base class for class node
#endif
#ifndef GRAPHDEF_H
#include "cl_graphdef.h"       // Type definitions that node uses
#endif
#ifndef ARCFLAGS_H
#include "cl_arcflags.h"       // ArcFlags class definition
#endif

class HashNodeList;
class ArcSet;
class AttributeSet;
class AvlPos;

typedef void(*VarArgFunRef)(...);
typedef int (*IntFunction)(...);

// Types of arc that are possible for a node.

typedef enum
{ inArcs, outArcs, inLevArcs, outLevArcs, invalidArcs } AdjList;


class GraphNode : public RigiObject
{
    NodeId        nId;         // database id
    NodeLabel     nLa;         // String label for the node

    unsigned int  info;        // for type(16), level(15), and loaded(1)

    void*         active;      // Pointer to display-specific info
    AttributeSet* attributes;  // Pointer to attribute set

    NDC_Point2d   pos;         // Normalized location of the node

    ArcSet *srcarcs;           // outgoing arcs (all types)
    ArcSet *dstarcs;           // incoming arcs (all types)

public:

    // Constructor and destructor for the class

    GraphNode(NodeLabel);
    ~GraphNode();

    // Access functions for the various instance variables

    void      SetId(NodeId);
    void      SetLabel(NodeLabel);
    void      SetType(NodeType);
    void      SetLevel();
    void      SetLevel(int);
    void      SetLoaded(int);

    void      SetActiveNode(void* ptr)     { active = ptr; }
    void      SetNDCPosition(NDC_Point2d);

    NodeId    GetId() const		   { return nId; }
    NodeLabel GetLabel()                   { return nLa; }
    NodeType  GetType()                    { return info>>16; }
    int       GetLevel()                   { return (info&0xffff)>>1; }
    RigiBool  GetLoaded()                  { return info&1; }

    void*       GetActiveNode()            { return active; }
    NDC_Point2d GetNDCPosition()           { return pos; }

    // Functions to get Predecessor and Successor sets for the node

    void GetPredecessor( HashNodeList& );
    void GetSuccessor( HashNodeList& );
    void GetPredecessor( HashNodeList&,AdjList,ArcType );
    void GetSuccessor( HashNodeList&,AdjList,ArcType );

    int   SetAttr( char*key, char*value );
    char* GetAttr( char*key );
    int   RemoveAttr( char*key );
    char* FirstAttr(AvlPos*&);
    char* NextAttr(AvlPos*&);
    int   NumAttrs();
    int   PrintAttrs();

    // Functions to add and delete arcs

    void InsertArc( GraphArc* );
    GraphArc* RemoveArc( GraphArc* );
    GraphArc* FindArc( GraphNode*, AdjList );
    GraphArc* FindArc( GraphNode*, AdjList, ArcType );

    GraphArc* FirstArc();
    GraphArc* FirstArc( const AdjList );

    GraphArc* FirstAdj( AvlPos*&, const AdjList );
    GraphArc* NextAdj( AvlPos*&, const AdjList );
    GraphArc* FirstArc( AvlPos*&, ArcType, ArcDirection &);
    GraphArc* NextArc(  AvlPos*&, ArcType, ArcDirection &);
    int       ArcNum( ArcDirection = ARC_INOUT );
    int       ArcNum( ArcType, ArcDirection = ARC_INOUT );

    void TreeDFS(ArcFlags &, ArcDirection, VarArgFunRef, void*,
                                    VarArgFunRef, void*,
                                    IntFunction, void* );
    void TreeBFS(ArcFlags &, ArcDirection, VarArgFunRef, void*,
                                    VarArgFunRef, void*,
                                    IntFunction, void* );
    void TreeDFS(AdjList, ArcType, VarArgFunRef, void*,
                                    VarArgFunRef, void*,
                                    IntFunction, void* );
    void TreeBFS(AdjList, ArcType, VarArgFunRef, void*,
                                    VarArgFunRef, void*,
                                    IntFunction, void* );
    void TreeDFS(ArcType, ArcDirection, VarArgFunRef, void*,
                                    VarArgFunRef, void*,
                                    IntFunction, void* );
    void TreeBFS(ArcType, ArcDirection, VarArgFunRef, void*,
                                    VarArgFunRef, void*,
                                    IntFunction, void* );
    void TreeDFS(AdjList, VarArgFunRef, void*, VarArgFunRef, void*);
    void TreeBFS(AdjList, VarArgFunRef, void*, VarArgFunRef, void*);

    // Copy the info from this node into the given node except the Id, and the
    // arc information. It only copies the immediate instance variables, none
    // of the lists or sets involved

    void Copy( GraphNode* );

    // Access function that determines whether or not the label of this node
    // fits the regular expression given.

    RigiBool GrepNodeLabel(char* regexp);

    //############################################################
    // This is the stuff needed for Craigs RigiObject class. First
    // the operators == and = for equality testing and assignment.
    //############################################################

    RigiBool operator==(const GraphNode& n) const;
    void operator=( GraphNode& n );

    // Now the routines required by the RigiObject class and any subclasses

    virtual int Printout(int) const;
    virtual unsigned int Hash() const;                 
    virtual RigiBool isEqual(void* a) const;
    virtual void* Duplicate() { return (NULL); } // not implemented
    virtual void* Create() { return (NULL); } // not implemented
    virtual void  Delete_type(ObjectPtr) {} // not implemented
};

#endif
