//
// File:    rigiwindow.h
// Author:  Brian Corrie
// Date:    January 16, 1991
// Project: Rigi V4.0
//

// $Author: kenw $
// $Date: 1998-06-30 23:01:56 $
// $Source: /root/cvs/Rigi/rigiedit/cl_rigiwindow.h,v $
// $Locker:  $
// $State: Exp $

//------------------------------------------------------------------------

#ifndef RIGIWINDOW_H
#define RIGIWINDOW_H

#ifndef CHARTYPEH
#include "adt/chartype.h"	// Include the CharType class
#endif

#ifndef LIST_H
#include "cl_list.h"            // Base class info for item lists
#endif
#ifndef ARCFLAGS_H
#include "cl_arcflags.h"        // ArcFlags class definition
#endif
#ifndef GRAPHDEF_H
#include "cl_graphdef.h"        // Point2d
#endif
#ifndef RIGIDEF_H
#include "cl_rigidef.h"         // Definition for Rigiwindow type stuff
#endif
#ifndef BASEWINDOW_H
#include "cl_base_window.h"
#endif
#ifndef FILTERSET_H
#include "cl_filterset.h"
#endif

class BaseWindow;
class NodeSet;
class RigiNodeSet;
class AvlPos;

// hints for faster drawing in a RigiWindow
extern const int DRAWACTIVE, DRAWNORMAL, DRAWNONAME, DRAWNAME;
extern const int DRAWICON, DRAWBOX, DRAWMOVE;

class RigiWindowList : public ItemList
{
public:

    RigiWindowList() {}
    ~RigiWindowList() {}
    RigiWindowList(ItemList &) {}

    // Routines provided by this class.

    void Insert( RigiWindow* wp ) { ItemList::Insert(wp); }
    RigiWindow* Remove( RigiWindow* wp )
        { return (RigiWindow *)ItemList::Remove( wp ); }
    RigiWindow* Find(RigiWindow *wp)
        { return (RigiWindow *)ItemList::Find(wp); }
    RigiWindow* First() { return (RigiWindow *)ItemList::First(); }
    RigiWindow* Next()  { return (RigiWindow *)ItemList::Next(); }
};

// Rigi window class that is generic across window systems (?).
// It contains a node list and arc list.

class RigiWindow
{
private:

    // General instance variables that should enable us to get rid of the
    // different types of windows and maintain only one type. If restrictions
    // are required then this class should be subclassed. The instance
    // variables node, type, projdepth, and overdepth should be rendundant.

    RigiWinType   type;             // The type of window it is.
    NodeSet*      nlist;            // Parent nodes for the window.
    ArcFlags      outarctypes;      // Out arctypes to traverse from the parents
    ArcFlags      inarctypes;       // In arc types to traverse from the parents
    int           depth;            // Depth to follow arcs to ( < 0 == infinte)
    char          inclusive;        // Include traversed nodes in the window?

    FilterSet*    nodeFilters;      // node filters
    FilterSet*    arcFilters;       // arc filters

    Point2d       canOrig;          // Location of the canvas in world coords
    float         xscale;           // horizontal scaling
    float         yscale;           // vertical scaling
    float         scale;            // for drawing objects (like nodes)

    int           total_arcs;       // total number of arcs in window
    int           filtered_nodes;   // Number of filtered nodes in window
    int           filtered_arcs;    // Number of filtered arcs in window

    CharType      arc_message;
    CharType      filter_message;
    char*         conditionstring;

    int  windownum;                 // Window number (used for scripting only)
    CharType windowname;            // Title of the window

    RigiNodeSet *nodeset;           // Set of nodes in this window

    BaseWindow* window;             // Pointer to a generic window

    void InitScratch( GraphNode* );
    void InitOverview( GraphNode* );
    void InitLevel( GraphNode* );
    void InitProjection( NodeSet& );
    void InitGeneral( NodeSet& list, ArcFlags& intypes, ArcFlags& outtypes,
                    int incl, int proj );
    void InitInstance( RigiWinType, int xloc, int yloc );
    void InitWindow( GraphNode*, int, int );
    void InitType( NodeSet&, RigiWinType );
    void ComputeScales();           // Compute scales so all nodes fit in window
    void DrawAll();                 // Draw all rigiobjects of the window
    void CenterScrollbars(int x, int y);    // position win & center scrollbars
    void GetNodeInWindow(Point *origin);

public:

    RigiWindow( NodeSet&, ArcFlags&, ArcFlags&, int, int, int, int,
                char *label = NULL );
    RigiWindow( NodeSet&, ArcFlags&, ArcDirection, int, int, int, int,
                char *label = NULL );
    RigiWindow( GraphNode*,RigiWinType,int,int,int,int );
    RigiWindow( GraphNode*,RigiWinType,int,int,int,int,int );
    RigiWindow( NodeSet&,RigiWinType,int,int,int,int,int );

    ~RigiWindow();

    // Access to the parent information

    GraphNode* GetNode();
    void       SetNode( GraphNode* );
    int        NumParents();
    GraphNode* NodeIsParent( NodeId );
    void       RemoveParent(NodeId);
    GraphNode* FirstParent( AvlPos*& );
    GraphNode* NextParent( AvlPos*& );

    // Destroy all the rigi stuff and recreate it

    void Empty();
    void Update();

    // Get the extents of all the nodes

    void GetWorldExtents( Point*, Point*, int );

    // Get and set the location of the window in World coordinate space. This
    // sets the location of the upper left corner of where this window looks
    // in on the larger space that exists in the Rigi editor.

    Point2d GetLocation()               { return canOrig; }
    void    SetLocation(Point2d p)      { canOrig = p; }

    // Scaling methods for RigiWindows.

    float GetScale()                    { return scale; }
    float GetXScale()                   { return xscale; }
    float GetYScale()                   { return yscale; }

    void  SetScale( float s )           { scale = s; UpdateLabel(); }
    void  SetXScale( float s )          { xscale = s; UpdateLabel(); }
    void  SetYScale( float s )          { yscale = s; UpdateLabel(); }

    void  ScaleToBox( Point, Point );
    void  ScaleToView();
    void  ScaleSelect();
    void  ScaleFactor();
    void  ScaleUnscale();

    // General access methods.

    RigiWinType GetType()               { return type; }
    void SetType( RigiWinType t )       { type = t; }

    void UpdateLabel( char *label = NULL, char *condition = NULL );
    int  Depth() const                  { return depth; }
    void SetDepth( int d )              { depth = d; }

    // RigiNode set iterators and access methods.

    int       NodeNumber();
    int       InsertNode( RigiNode* );
    RigiNode* RemoveNode( RigiNode* );
    RigiNode* FindNode( RigiNode* );
    RigiNode* FirstNode( AvlPos*& );
    RigiNode* NextNode( AvlPos*& );
    RigiNode* FindRigiNode( GraphNode* );

    void      CreateRigiArcs( RigiNode* );
    void      CreateRigiArcs( RigiNodeList* );

    // new methods added for improved window stats
    int	      NodeNumber( int type );
    int       ShowedNodeNumber();
    int       ShowedNodeNumber( int type );
    int	      ArcNumber( int type );
    int	      ShowedArcNumber();
    int       ShowedArcNumber( int type );
        
    RigiArc* FindRigiArc( GraphArc* );
    int      ArcNumber()                 { return total_arcs; }

    // Conversion functions between Window, World, and Normalized Device
    // Coordinate spaces. Window coordinates are relative to this windows
    // origin (Set with the method SetLocation). World coordinates have the
    // absolute zero point 0,0,0 and a maximum point defined and hidden in
    // cl_rigiwindow.c. Normalized Device Coordinates (NDC) range from 0 to
    // 1, and are determined by the space defined by the World Coordinate
    // system.

    // Actually "window" in these functions refers to the canvas.
    // Eventually, their names will be changed.  JGM  Dec 18, 95.

    Point       NDC2World( NDC_Point );
    Point       NDC2World( NDC_Point2d );
    Point2d     NDC2World2d( NDC_Point2d );
    NDC_Point   World2NDC( Point );
    NDC_Point2d World2NDC2d( Point );
    NDC_Point2d World2NDC2d( Point2d );
    Point       World2Canvas( Point );
    Point       World2Canvas( Point2d );
    Point       Canvas2World( Point );


    RigiNode* FindNode( Point );
    RigiArc* FindArc( Point );
    int CheckNode( RigiNode* );
    int CheckArc( RigiArc* );

    int GetArcType( RigiArc* );

    // Window methods

    BaseWindow* Window()	             { return window; }

    // Funcs to translate the window origin to the canvas
    // origin (and vice versa).  These functions treat
    // the window as an unscaled offset from the canvas
    // origin.

    Point2d     Window2Canvas(Point2d p) { return window->Window2Canvas(p); }
    Point2d     Canvas2Window(Point2d p) { return window->Canvas2Window(p); }
    Point       Window2Canvas(Point p)   { return window->Window2Canvas(p); }
    Point       Canvas2Window(Point p)   { return window->Canvas2Window(p); }

    void Clear()                         { window->Clear(); }
    void DrawSynch( RigiBool i )         { window->DrawSynch(i); }
    void BunchNode( RigiBool i )         { window->BunchRect(i); }
    void BunchArc( RigiBool i )          { window->BunchLine(i); }
    void Batch( RigiBool i )             { window->Batch(i); }
    void BackStore( RigiBool i )         { window->BackStore(i); }
    RigiBool BackStore()                 { return window->BackStore(); }
    void Drawing( RigiBool i )           { window->Drawing(i); }
    void Hide()                          { window->Hide(); }
    void Show()                          { window->Show(); }
    void RestoreWindow()                 { window->RestoreWindow(); }
    void Restore(int x, int y, int w, int h)  { window->Restore( x,y, w,h ); }
    void Dimensions( int*x,int*y, int*w,int*h );

    void NodeExtents( RigiNode*, Point*, Point* );
    int  WindowNodeExtents( RigiNode*, Point*, Point* );
    Point InArcPoint( RigiNode* );
    Point OutArcPoint( RigiNode* );
    void DrawNode( RigiNode* );
    void DrawIcon( Point p, int size, int type, int active );
    void DrawSquare( Point sp, int size, int thick, int type );
    void DrawBox( Point sp, Point dp, int thick, int type, int dpyOnly );
    void DrawArc( RigiArc* );
    void DrawLine( Point sp, Point dp, int width, int style );
    void DrawString( char* str, Point p );
    void DrawStore();
    void RigiDisplay();
    void RigiDisplay( int, Point, Point );
    void DestroyChildWindows();

    void UpdateArcMessage(char* str);
    void UpdateFilterMessage(); 
    void SetNumFilteredNodes( int num );
    int  GetNumFilteredNodes();
    void SetNumFilteredArcs( int num );
    int  GetNumFilteredArcs();

    FilterSet *GetArcTypeFilters() {return arcFilters;}
    FilterSet *GetNodeTypeFilters() {return nodeFilters;}
    void InheritTypeFilters(const RigiWindow *baseWin);

    // Archive methods
    int WriteArchive( FILE* fp );
    int ReadArchive( FILE* fp );

    // SCRIPT10: add identifier numbers to the windows
    int GetWindowNumber()           { return windownum; }
    void SetWindowNumber( int num ) { windownum = num; } 
};


#endif


