
#ifndef EDITUTIL_H
#include "editutil.h"
#endif
#ifndef RIGIWINDOW_H
#include "cl_rigiwindow.h"
#endif
#ifndef GRAPHNODE_H
#include "cl_graphnode.h"
#endif
#ifndef GRAPHARC_H
#include "cl_grapharc.h"
#endif
#ifndef GRAPHNODELIST_H
#include "cl_graphnodelist.h"
#endif
#ifndef GRAPHARCLIST_H
#include "cl_grapharclist.h"
#endif
#ifndef RIGINODE_H
#include "cl_riginode.h"
#endif
#ifndef GRAPHNODESET_H
#include "cl_graphnodeset.h"
#endif

#ifndef GRAPHUTIL_H
#include "graphutil.h"
#endif
#ifndef RIGIERR_H
#include "rigierr.h"
#endif
#ifndef RIGIUTIL_H
#include "rigiutil.h"
#endif
#ifndef RIGIINIT_H
#include "rigi_init.h"
#endif
#ifndef RIGI_H
#include "rigi.h"
#endif

#ifndef RIGISTATE_H
#include "cl_rigistate.h"       // Global instance of editor's state
#endif
#ifndef GLOBAL_H
#include "global.h"
#endif
#ifndef DOMAIN_H
#include "cl_domain.h"          // NumArcTypes() 
#endif
#ifndef GRAPHDEF_H
#include "cl_graphdef.h"
#endif
#ifndef EDIT_H
#include "edit.h"
#endif

static void GetWindowsWithNode(GraphNode *graphNode, SimpleList &rwList) {
    RigiWindow *rw = editor->FirstWindow();
    while (rw != NULL) {
        if (rw->FindRigiNode(graphNode) != NULL)
            rwList.Insert((void*)rw);
        rw= editor->NextWindow();
    }
}

static void UpdateWindows(SimpleList &rwList) {
    RigiWindow *rw;
    while ((rw = (RigiWindow*)rwList.RemoveFirst()) != NULL)
        rw->Update();
}

// updates windows without deselecting nodes, arcs or defiltering nodes.
// can be called when adding nodes or arcs.
void EditUpdateAllWindows() {
    SimpleList rwList;
    RigiWindow *rw = editor->FirstWindow();
    while (rw != NULL) {
        rwList.Insert((void*)rw);
        rw= editor->NextWindow();
    }
    UpdateWindows(rwList);
}

void EditClear() {
    NodeList cliplist;

    // Get all destination nodes of all level arcs that leave the clipnode.
    // These are the roots of the trees to be cleaned up.

    edit_get_cliplist( cliplist );

    // Delete all of the nodes in the hierarchies defined by the cliplist
    edit_delete_all( cliplist );

    // Update the clip board window

    RigiUpdateClipBoard();
}

void EditCut() {
    NodeList locallist;

    Mess( "EditCut: Cleaning up clipboard" );
    EditClear();

    Mess("EditCut: %d selected nodes", editor->NumSelectedNodes() );
    editor->GetSelectionList( locallist );
    editor->EmptySelection();

    // Cleanup the stuff that has to do with the set of nodes we are cutting.
    // This includes deleting all the rigi info involved, and cutting all
    // the level arcs that connect the clip stuff to the clipnode

    Mess("EditCut: Update list to cut");
    edit_update_list_to_cut( locallist );

    Mess("EditCut: disconnect parent");
    GraphNode* parent = RutilActiveWindow()->GetNode();
    edit_disconnect_node( parent, locallist );

    Mess("EditCut: Adding to the clipnode");
    edit_add_clipnode( locallist );

    RigiUpdateClipBoard();
}

void EditCopy( RigiWindow* window ) {
    NodeSet copynodes, newnodes;
    NodeList duplist,locallist;

    // Copy the selection list into the local list.

    editor->GetSelectionList( locallist );
    editor->EmptySelection();

    // Clean out the clipboard if anything is there

    EditClear();

    // Get the ids of the nodes to be copied.

    edit_get_ids( locallist, copynodes );

    // Duplicate all the nodes that are below the selected nodes.

    edit_duplicate_nodes( locallist, copynodes, duplist );

    // Disconnect the duplicated copies from the parent.

    GraphNode* parent = window->GetNode();
    edit_disconnect_node( parent, duplist );

    // Get the ids of the new nodes

    edit_get_ids( duplist, newnodes );

    // Get rid of all the rigi stuff associated with these guys.

    edit_update_list( duplist, newnodes );

    // Add the (hopefully de-rigified) nodes to the clipboard and update it

    edit_add_clipnode( duplist );
    RigiUpdateClipBoard();
}

void EditPaste( RigiWindow* window ) {
    NodeList cliplist;

    // Get the list of nodes that are in the clipboard

    Mess("Paste - Getting the cliplist" );
    edit_get_cliplist( cliplist );

    // Paste the ids into the window

    Mess("Paste - Pasting the nodes");
    edit_paste_nodes( cliplist, window );
}


int EditCollapse( Point p, char *name, int nodetype ) {
    NodeSet collocallist, externallist, compnode;
    ArcList arcdellist, comparcs, incomparcs, outcomparcs;
    GraphNode *np; AvlPos *P;
    ArcDirection arcout = ARC_OUT;

    int lt = editor->LevelArc();

    RigiWindow *rw = RutilActiveWindow();
    if ( rw==NULL )
        return RigiError(ERRORFLAG, "NULL active window" );

    // expand is allowed only in those windows where 
    // the children of a collapsed node are NOT shown.


    GraphNode* parent = rw->GetNode();
    if (parent==NULL)
        return RigiError(ERRORFLAG, "NULL parent in EditCollapse()");

    // MJW 931018
    int parentlevel = (parent==NULL) ? 0 : parent->GetLevel();

    // Copy the selection list into the local list.
    // Separate all nodes that are in the active window from the ones
    // that are in some other windows.
    // Check if node's levels are OK so the k2 structure will be preserved.

    GraphNode* node = editor->FirstSelectedNode(P);
    while ( node ){
        extern RigiBool K2_CHECK;
        RigiNode *winnode = rw->FindRigiNode( node );
        if ( winnode == NULL ) {
            if ( !externallist.Find(node->GetId()) )
                externallist.Add( node );

            node->SetLevel();
            if (K2_CHECK) if ( node->GetLevel() != parentlevel+2 )
                return RigiError(ERRORFLAG, "Collapse is not allowed for these nodes");
        } else {
            NDC_Point2d ndcp;
            if (collocallist.Find( node->GetId() ) == NULL )
                collocallist.Add( node );
            node->SetLevel();
            ndcp = rw->World2NDC2d( winnode->GetPosition() );
            node->SetNDCPosition(ndcp);
            if (K2_CHECK) if ( node->GetLevel() != parentlevel+1 )
                return RigiError(ERRORFLAG, "Collapse is not allowed for these nodes");
        }
        node = editor->NextSelectedNode(P);
    }

    editor->EmptySelection();

    // Delete all the rigi stuff associated with these nodes, get the world
    // location of the first node, and collect a list of composite arcs to
    // and from nodes in the local list. This list is used to update the
    // composite arcs to the new collapse node.

    
    for ( np=collocallist.First(P); np; np=collocallist.Next(P)) {
        edit_delete_rigi_nodes( np );
        edit_get_comp_arcs(np,collocallist,incomparcs,outcomparcs);
    }

    // Collect a list of arcs to delete from the parent to the collapse nodes

    GraphArc *ga = (parent==NULL) ? NULL : parent->FirstArc(P, lt, arcout);

    while ( ga ) {
        GraphNode* np = ga->GetDst();
        if ( collocallist.Find( np->GetId() ) != NULL )
            arcdellist.Insert( ga );

        ga = parent->NextArc(P, lt, arcout);
    }

    // Delete all the level arcs from the parent to the collapse nodes

    for ( ga = arcdellist.First(); ga; ga = arcdellist.Next() )
        GutilDeleteArc( ga );

    // Create and connect the collapse node to the parent

    if (!name) name = "collapse";
    GraphNode* colnode = GutilCreateNode( parent, name, nodetype );

    // Set up the composite arc from other nodes to the collapse node

    GraphArc* ap;
    for ( ap = incomparcs.First(); ap; ap = incomparcs.Next() ) {
        GraphNode* src = ap->GetSrc();
        if (src==NULL) mayday("unexpected NULL src in EditCollapse()");

        // If we haven't already created a composite arc for this external node,
        // create one and copy the contributor ids. Otherwise just copy the
        // contributor ids to the composite arc we have already created.

        if ( compnode.Find( src->GetId() ) == NULL ) {
            // Add the node to the list

            compnode.Add( src );

            // Create a new composite arc

            GraphArc*comparc=new GraphArc( src,colnode,editor->CompositeArc() );
            src->InsertArc( comparc );
            colnode->InsertArc( comparc );

            for ( int i=0; i<ap->NumContributorIds(); i++ ) {
                comparc->AddContributorId( ap->GetContributorId( i ) );
            }

            // Add the composite arc to the list we have created so far.

            comparcs.Insert( comparc );

            // Create the editor arc
            GutilAddArc( comparc );
        } else {
            int gotarc = 0;
            GraphArc* ca;
            for ( ca = comparcs.First(); ca; ca = comparcs.Next() )
                if ( ca->GetSrc()->GetId() == src->GetId() ) {
                    for ( int i=0; i<ap->NumContributorIds(); i++ )
                        ca->AddContributorId( ap->GetContributorId( i ) );
                    gotarc = 1;
                    break;
                }
            if (!gotarc)
                mayday("****Error: EditCollapse - Failed to find composite arc");
        }
    }

    // Set up the composite arc from a collapse node to an external node.

    compnode.Empty();
    for ( ap = outcomparcs.First(); ap; ap = outcomparcs.Next() ) {
        GraphNode* dst = ap->GetDst();

        // If we haven't already created a composite arc for this external node,
        // create one and copy the contributor ids. Otherwise just copy the
        // contributor ids to the composite arc we have already created.

        if ( !compnode.Find(dst->GetId()) ){
            // Add the node to the list

            compnode.Add( dst );

            // Create a new composite arc

            GraphArc *ca = new GraphArc( colnode,dst,editor->CompositeArc() );
            dst->InsertArc( ca );
            colnode->InsertArc( ca );

            for ( int i=0; i<ap->NumContributorIds(); i++ )
                ca->AddContributorId( ap->GetContributorId( i ) );

            // Add the composite arc to the list we have created so far.

            comparcs.Insert( ca );

            // Create the editor arc
            GutilAddArc( ca );
        } else {
            int gotarc = 0;
            GraphArc* ca;
            for ( ca = comparcs.First(); ca; ca = comparcs.Next() )
                if ( ca->GetDst()->GetId() == dst->GetId() ) {
                    for ( int i=0; i<ap->NumContributorIds(); i++ )
                        ca->AddContributorId( ap->GetContributorId( i ) );
                    gotarc = 1;
                    break;
                }
            if (!gotarc) mayday("EditCollapse - failed to find comp arc");
        }
    }

    // Connect all the local-collapse nodes to the new node

    for ( np=collocallist.First(P); np; np=collocallist.Next(P)) {
        // Create a new level arc

        GraphArc* ap = new GraphArc( colnode,np,lt );
        colnode->InsertArc( ap );
        np->InsertArc( ap );
        colnode->SetLevel();
        np->SetLevel();

        // Create the editor arc
        GutilAddArc( ap );
    }

    // Connect all the external-collapse nodes to the new node

    for ( np=externallist.First(P); np; np=externallist.Next(P) ) {
        // Create a new level arc

        GraphArc* ap = new GraphArc( colnode,np,lt );
        colnode->InsertArc( ap );
        np->InsertArc( ap );

        // Create the editor arc
        GutilAddArc( ap );
    }

    RigiNode* rn = new RigiNode( colnode, rw );
    Point2d pos; pos.x = p.x; pos.y = p.y;
    rn->SetPosition( pos );
    rn->Draw(); // MJW
    rw->InsertNode( rn );
    rw->CreateRigiArcs( rn );

    // MJW: select the newly-created riginode
    editor->SelectNode(rn);
    return colnode->GetId();
}

int EditExpand() {
    NodeList explist;
    ArcList arcdellist;
    SimpleList rwList;
    ArcDirection dir = ARC_OUT;
    int id;

    int lt = editor->LevelArc();

    RigiWindow *rw = RutilActiveWindow();
    if (rw==NULL)
        return RigiError(ERRORFLAG, "NULL active window");


    if ( editor->NumSelectedNodes() != 1 )
        return RigiError(ERRORFLAG, "Exactly one node must be selected");


    AvlPos *P;
    GraphNode *expnode = editor->FirstSelectedNode(P);

    RigiNode* rn = rw->FindRigiNode( expnode );
    if (rn == NULL)
        return RigiError(ERRORFLAG, "Exactly one node must be selected");

    // Check to make sure the node being expanded is not a leaf node.

    GraphArc *ga;
    if ((ga = expnode->FirstArc(P, lt, dir)) == NULL)
        return RigiError(ERRORFLAG,
            "Node %s is a leaf node, cannot expand", expnode->GetLabel());

    // expand is allowed only in those windows where 
    // the children of a collapsed node are NOT shown.

    while (ga != NULL) {
        if (rw->FindRigiNode(ga->GetDst()) != NULL)
            return RigiError(ERRORFLAG, "Expand is not allowed in this window");
        ga = expnode->NextArc(P, lt, dir);
    }

    // Get the node to expand, and make sure its entire subtree is in memory

    id = expnode->GetId();
    editor->EmptySelection();

    // Erase and delete the riginode, and its associated rigiarcs

    rn->Draw();
    for (RigiArc *ra = rn->FirstArc(P); ra; ra = rn->FirstArc(P) ) {
        ra->Draw();
        delete ra;  // destructor will remove from riginodes' lists
    }
    rw->RemoveNode(rn);
    delete rn;

    // Compile the list of nodes that the expand node is connected to
    ga = expnode->FirstArc(P, lt, dir);
    while ( ga != NULL ) {
        GraphNode* np = ga->GetDst();
        if ( np->ArcNum(lt, ARC_IN) == 1) explist.Insert( np );
        ga = expnode->NextArc(P, lt, dir);
    }

    // Compile a list of arcs to connect to children of the expand node.
    // These are all the incoming levelarcs, and all others (in and out).

    int ntypes = RinitGetDomain()->NumArcTypes();

    for(dir = ARC_IN; ; dir = ARC_OUT) {
        for ( int arctype=0; arctype<ntypes; arctype++ ) {
            if (dir == ARC_OUT && arctype == lt) continue;

            for (ga = expnode->FirstArc(P, arctype, dir); ga;
                ga = expnode->NextArc(P,arctype,dir) ) {
                GraphNode* othernode = NULL;

                if ( ga->GetSrc() == expnode )
                    othernode = ga->GetDst();
                else if ( ga->GetDst() == expnode )
                    othernode = ga->GetSrc();
                else
                    mayday("Couldn't find the other node" );

                // do not add self-arcs twice
                if (othernode==expnode && dir==ARC_OUT) continue;
                arcdellist.Insert( ga );

                if ( dir!=ARC_IN || arctype!=lt ) continue;

                for ( GraphNode *np=explist.First(); np; np=explist.Next() ) {
                    GraphArc* ap = new GraphArc( othernode,np,ga->GetType() );
                    np->InsertArc( ap );
                    othernode->InsertArc( ap );

                    // Create the editor arc
                    GutilAddArc( ap );
                }

            }
        }
        if (dir == ARC_OUT) break;
    }

    // Delete the arcs to and from the expand node

    for ( ga = arcdellist.First(); ga; ga = arcdellist.Next() )
        GutilDeleteArc( ga );

    // Delete outLevArcs for expand node since they are not on deletelist

    ga = expnode->FirstArc( P, lt, dir);
    while ( ga != NULL ) {
        GutilDeleteArc( ga );
        ga = expnode->FirstArc( P, lt, dir);
    }

    GetWindowsWithNode(expnode, rwList);

    // Delete the node and any windows which are children to the node.

    GutilDeleteNode(expnode);

    // It is possible that the node is both parent and child ...
    // Check before redrawing the window (it may have been destroyed).
    // This is an ugly hack.  JGM 25 Oct 96.

    if (editor->FindWindow(rw) == NULL) return 1;

    // Set the new level for the expanded nodes and
    // draw the expanded nodes.
    RigiNodeList rnl;
    for ( GraphNode *np = explist.First(); np; np = explist.Next() ) {
        np->SetLevel();
        // create rigi nodes and arcs
        RigiNode* rn = new RigiNode(np,rw);
        // Get the NDC position of the node and transform to local coords.
        NDC_Point2d ndcp = np->GetNDCPosition();
        Point2d worldp = rw->NDC2World2d( ndcp );
        rn->SetPosition( worldp );
        rn->Draw(); // MJW
        editor->SelectNode(rn);
        rw->InsertNode(rn);
        rnl.Insert(rn);
    } 

    // create rigi arcs for each node
    rw->CreateRigiArcs(&rnl);
    return 1;
}

int EditSelectAll() {
    RigiWindow *rw = RutilActiveWindow();
    if (rw==NULL)
        return RigiError(ERRORFLAG, "NULL active window");

    editor->EmptySelection();
    AvlPos *P;
    for ( RigiNode* rn = rw->FirstNode(P); rn; rn = rw->NextNode(P) )
        editor->SelectNode( rn );
    return 1;
}
